"""Exceptions for Censys."""
from typing import Dict, Optional, Type


class CensysException(Exception):
    """Base Exception for Censys."""


class CensysCLIException(CensysException):
    """Exception raised when the CLI is passed invalid arguments."""


class CensysAPIException(CensysException):
    """Base Exception for Censys APIs."""

    def __init__(
        self,
        status_code: int,
        message: str,
        body: Optional[str] = None,
        const: Optional[str] = None,
        error_code: Optional[int] = None,
        details: Optional[str] = None,
    ):
        """Inits CensysAPIException.

        Args:
            status_code (int): HTTP status code.
            message (str): HTTP message.
            body (str): Optional; HTTP body.
            const (str): Optional; Constant for manual errors.
            error_code (int): Optional; Error code.
            details (str): Optional; Additional details.
        """
        self.status_code = status_code
        self.message = message
        self.body = body
        self.const = const
        self.error_code = error_code
        self.details = details
        super().__init__(self.message)


class CensysSearchException(CensysAPIException):
    """Base Exception for the Censys search API."""

    def __repr__(self) -> str:
        """Representation of CensysSearchException.

        Returns:
            str: Printable representation.
        """
        return f"{self.status_code} ({self.const}): {self.message or self.body}"

    __str__ = __repr__


class CensysAsmException(CensysAPIException):
    """Base Exception for the Censys ASM API."""

    def __repr__(self) -> str:
        """Representation of CensysAsmException.

        Returns:
            str: Printable representation.
        """
        return (
            f"{self.status_code} (Error Code: {self.error_code}), "
            f"{self.message}. {self.details}"
        )

    __str__ = __repr__


class CensysMissingApiKeyException(CensysAsmException):
    """Exception raised when there is no provided ASM API key."""


class CensysRateLimitExceededException(CensysSearchException):
    """Exception raised when your Censys rate limit has been exceeded."""


class CensysNotFoundException(CensysSearchException):
    """Exception raised when the resource requested is not found."""


class CensysUnauthorizedException(CensysSearchException):
    """Exception raised when you doesn't have access to the requested resource."""


class CensysJSONDecodeException(CensysSearchException):
    """Exception raised when the resource requested is not valid JSON."""


class CensysInvalidRequestException(CensysAsmException):
    """Exception raised when the HTTP request is invalid."""


class CensysInvalidAuthTokenException(CensysAsmException):
    """Exception raised when the auth token is invalid."""


class CensysAsmUnauthorizedException(CensysAsmException):
    """Exception raised when the ASM API is unauthorized."""


class CensysInvalidAPIKeyException(CensysAsmException):
    """Exception raised when the ASM API key is invalid."""


class CensysTooManyRequestsException(CensysAsmException):
    """Exception raised when the allowed requests bandwidth is exceeded."""


class CensysAppDownForMaintenanceException(CensysAsmException):
    """Exception raised when the ASM API is down for maintenance."""


class CensysInvalidSeedDataException(CensysAsmException):
    """Exception raised when the seed data is invalid."""


class CensysAssociatedAssetsThresholdWarningException(CensysAsmException):
    """Exception raised when the asset count is within the warning threshold."""


class CensysTooManyInputNodesException(CensysAsmException):
    """Exception raised when there are too many input nodes."""


class CensysSeedNotFoundException(CensysAsmException):
    """Exception raised when the requested seed can not be found."""


class CensysNotASeedException(CensysAsmException):
    """Exception raised when the requested resource is not a seed."""


class CensysNeedConfirmationToRemoveParentSeedsException(CensysAsmException):
    """Exception raised when confirmation is needed to remove seeds with children."""


class CensysCannotRemoveNonExistentSeedsException(CensysAsmException):
    """Exception raised when trying to remove non existent seed nodes."""


class CensysCannotRemoveNonSeedsException(CensysAsmException):
    """Exception raised when trying to remove non seed nodes."""


class CensysInvalidSeedTypeException(CensysAsmException):
    """Exception raised when the seed type is invalid."""


class CensysInvalidLogbookCursorException(CensysAsmException):
    """Exception raised when the logbook cursor is invalid."""


class CensysTeamNotFoundException(CensysAsmException):
    """Exception raised when the team is not found."""


class CensysPageNumberOutOfRangeException(CensysAsmException):
    """Exception raised when the page number is out of range [1 - totalPages]."""


class CensysInvalidPageSizeException(CensysAsmException):
    """Exception raised when the page size is invalid."""


class CensysHostNotFoundException(CensysAsmException):
    """Exception raised when the requested host is not found."""


class CensysInvalidIPv4AddressException(CensysAsmException):
    """Exception raised when the IPv4 address is invalid."""


class CensysAssetExludedException(CensysAsmException):
    """Exception raised when the asset is excluded."""


class CensysInvalidCommentException(CensysAsmException):
    """Exception raised when the comment is invalid."""


class CensysCommentNotFoundException(CensysAsmException):
    """Exception raised when the requested comment is not found."""


class CensysSubdomainNotFoundException(CensysAsmException):
    """Exception raised when the requested subdomain is not found."""


class CensysInvalidColorException(CensysAsmException):
    """Exception raised when the specified color is invalid."""


class CensysTagHasTrailingOrLeadingWhitespaceException(CensysAsmException):
    """Exception raised when the specified tag has trailing or leading whitespace."""


class CensysTagIsEmptyStringException(CensysAsmException):
    """Exception raised when the specified tag is an empty string."""


class CensysTagLabelsDifferOnlyInCasingException(CensysAsmException):
    """Exception raised when the tag differs from an existing tag in only casing."""


class CensysTagLabelTooLongException(CensysAsmException):
    """Exception raised when the specified tag label is too long."""


class CensysTagColorTooLongException(CensysAsmException):
    """Exception raised when the specified tag color is too long."""


class CensysCannotCreateTagWithNewColorException(CensysAsmException):
    """Exception raised when the specified tag cannot be created with a new color."""


class CensysTagColorHasTrailingOrLeadingWhitespaceException(CensysAsmException):
    """Exception raised when the tag color has trailing or leading whitespace."""


class CensysCertificateNotFoundException(CensysAsmException):
    """Exception raised when the certificate is not found."""


class CensysDomainNotFoundException(CensysAsmException):
    """Exception raised when the domain is not found."""


class CensysInvalidCloudAssetDataException(CensysAsmException):
    """Exception raised when invalid cloud asset data is submitted."""


class CensysInvalidObjectStorageAssetIdentifierException(CensysAsmException):
    """Exception raised when object storage name is not a valid asset URL."""


class CensysInvalidObjectStorageAssetNotFoundException(CensysAsmException):
    """Exception raised when no object storage assets with given URL were found."""


class CensysBadJSONBodyException(CensysAsmException):
    """Exception raised when a bad JSON string is in the body."""


class CensysRiskNotFoundException(CensysAsmException):
    """Exception raised when no risks are found with given risk_id."""


class CensysInvalidDateException(CensysAsmException):
    """Exception raised when an invalid date is submitted."""


class CensysInvalidCloudException(CensysAsmException):
    """Exception raised when an invalid cloud is submitted."""


class CensysAssetNotFoundException(CensysAsmException):
    """Exception raised when the asset is not found."""


class CensysInvalidKeywordsInBodyException(CensysAsmException):
    """Exception raised when invalid keywords are in the body."""


class CensysSearchAPITimeoutException(CensysAsmException):
    """Exception raised when the search API times out."""


class CensysSearchAPIErrorException(CensysAsmException):
    """Exception raised when the search API returns an error."""


class CensysInternalServerErrorException(CensysAsmException):
    """Exception raised when the search API returns an error."""


class CensysInvalidCommentHTMLException(CensysAsmException):
    """Exception raised when the comment contains invalid HTML."""


class CensysInvalidSearchAPIResponseException(CensysAsmException):
    """Exception raised when the search API returns an error."""


class CensysTooSoonToResendInviteException(CensysAsmException):
    """Exception raised when it is too soon to resend the invite."""


class CensysExceptionMapper:
    """Map status code to Exception for the ASM and Search API."""

    ASM_EXCEPTIONS: Dict[int, Type[CensysAsmException]] = {
        10000: CensysMissingApiKeyException,
        10001: CensysInvalidAPIKeyException,
        10002: CensysInvalidAuthTokenException,
        10006: CensysAsmUnauthorizedException,
        10007: CensysInvalidSeedDataException,
        10008: CensysInvalidRequestException,
        10011: CensysCannotRemoveNonSeedsException,
        10012: CensysCannotRemoveNonExistentSeedsException,
        10013: CensysNeedConfirmationToRemoveParentSeedsException,
        10014: CensysSeedNotFoundException,
        10015: CensysNotASeedException,
        10016: CensysTooManyInputNodesException,
        10017: CensysAssociatedAssetsThresholdWarningException,
        10018: CensysHostNotFoundException,
        10019: CensysDomainNotFoundException,
        10020: CensysCertificateNotFoundException,
        10021: CensysInvalidIPv4AddressException,
        10022: CensysAssetExludedException,
        10025: CensysTagHasTrailingOrLeadingWhitespaceException,
        10026: CensysTagIsEmptyStringException,
        10027: CensysTagLabelsDifferOnlyInCasingException,
        10028: CensysTagLabelTooLongException,
        10029: CensysAppDownForMaintenanceException,
        10034: CensysTagColorTooLongException,
        10035: CensysCannotCreateTagWithNewColorException,
        10036: CensysTagColorHasTrailingOrLeadingWhitespaceException,
        10037: CensysInvalidColorException,
        10038: CensysInvalidSeedTypeException,
        10039: CensysTooManyRequestsException,
        10040: CensysInvalidLogbookCursorException,
        10045: CensysTeamNotFoundException,
        10050: CensysInvalidPageSizeException,
        10051: CensysPageNumberOutOfRangeException,
        10054: CensysInvalidCommentException,
        10055: CensysCommentNotFoundException,
        10057: CensysSubdomainNotFoundException,
        10059: CensysInvalidCloudAssetDataException,
        10060: CensysInvalidObjectStorageAssetIdentifierException,
        10061: CensysInvalidObjectStorageAssetNotFoundException,
        10067: CensysBadJSONBodyException,
        10073: CensysRiskNotFoundException,
        10078: CensysInvalidDateException,
        10082: CensysInvalidCloudException,
        10086: CensysAssetNotFoundException,
        10091: CensysInvalidKeywordsInBodyException,
        10096: CensysSearchAPITimeoutException,
        10097: CensysSearchAPIErrorException,
        10098: CensysInternalServerErrorException,
        10099: CensysInvalidCommentHTMLException,
        10107: CensysInvalidSearchAPIResponseException,
        10120: CensysTooSoonToResendInviteException,
    }
    """Map of status code to ASM Exception."""

    SEARCH_EXCEPTIONS: Dict[int, Type[CensysSearchException]] = {
        401: CensysUnauthorizedException,
        403: CensysUnauthorizedException,
        404: CensysNotFoundException,
        429: CensysRateLimitExceededException,
    }
    """Map of status code to Search Exception."""



"""Warns on deprecated class and functions."""
import functools
import warnings
from typing import Callable, Optional


class DeprecationDecorator:
    """Deprecation Decorator for classes and functions."""

    def __init__(self, message: Optional[str] = None):
        """Inits DeprecationDecorator.

        Args:
            message (str): Optional; Message to display to user.
        """
        self.message = message

    def __call__(self, func: Callable) -> Callable:
        """Wrapper function.

        Args:
            func (Callable): Function to wrap.

        Returns:
            Callable: Wrapped function.
        """

        @functools.wraps(func)
        def new_func(*args, **kwargs):
            # warnings.simplefilter("always", DeprecationWarning)
            warnings.warn(
                self.message or f"Call to deprecated function {func.__name__}.",
                category=DeprecationWarning,
                stacklevel=2,
            )
            # warnings.simplefilter("default", DeprecationWarning)
            return func(*args, **kwargs)

        return new_func



"""Interact with the Censys Seeds API."""
from typing import Optional

from .api import CensysAsmAPI

SEED_TYPES = ["IP_ADDRESS", "DOMAIN_NAME", "CIDR", "ASN"]


class Seeds(CensysAsmAPI):
    """Seeds API class."""

    base_path = "/v1/seeds"

    def get_seeds(
        self, seed_type: Optional[str] = None, label: Optional[str] = None
    ) -> dict:
        """Requests seed data.

        Args:
            seed_type (str):
                Optional; Seed type ['IP_ADDRESS', 'DOMAIN_NAME', 'CIDR', 'ASN'].
            label (str): Optional; Seed label.

        Returns:
            dict: Seed search results.
        """
        args = {}
        if seed_type:
            args["type"] = seed_type
        if label:
            args["label"] = label
        return self._get(self.base_path, args=args)["seeds"]

    def get_seed_by_id(self, seed_id: int) -> dict:
        """Requests seed data by ID.

        Args:
            seed_id (int): Seed ID to get.

        Returns:
            dict: Seed search result.
        """
        path = f"{self.base_path}/{seed_id}"

        return self._get(path)

    def add_seeds(self, seeds: list, force: Optional[bool] = None) -> dict:
        """Add seeds to the ASM platform.

        Args:
            seeds (list): List of seed objects to add.
            force (bool, optional): Forces replace operation.

        Returns:
            dict: Added seeds results.
        """
        data = {"seeds": seeds}
        args = {"force": force}

        return self._post(self.base_path, args=args, data=data)

    def replace_seeds_by_label(
        self, label: str, seeds: list, force: Optional[bool] = None
    ) -> dict:
        """Replace seeds in the ASM platform by label.

        Args:
            label (str): Label name to replace by.
            seeds (list): List of seed objects to add.
            force (bool): Optional; Forces replace operation.

        Returns:
            dict: Added and removed seeds results.
        """
        data = {"seeds": seeds}
        args = {"label": label, "force": force}

        return self._put(self.base_path, args=args, data=data)

    def delete_seeds_by_label(self, label: str) -> dict:
        """Delete seeds in the ASM platform by label.

        Args:
            label (str): Label name to delete by.

        Returns:
            dict: Delete results.
        """
        args = {"label": label}

        return self._delete(self.base_path, args=args)

    def delete_seed_by_id(self, seed_id: int) -> dict:
        """Delete a seed in the ASM platform by id.

        Args:
            seed_id (int): Seed ID to delete by.

        Returns:
            dict: Delete results.
        """
        path = f"{self.base_path}/{seed_id}"

        return self._delete(path)



"""Censys view CLI."""
import argparse
import sys
import webbrowser

from censys.cli.utils import V2_INDEXES, console, valid_datetime_type, write_file
from censys.search import SearchClient
from censys.search.v2.api import CensysSearchAPIv2


def cli_view(args: argparse.Namespace):
    """Search subcommand.

    Args:
        args (Namespace): Argparse Namespace.
    """
    if args.open:
        webbrowser.open(
            f"https://search.censys.io/{args.index_type}/{args.document_id}"
        )
        sys.exit(0)

    censys_args = {}

    if args.api_id:
        censys_args["api_id"] = args.api_id

    if args.api_secret:
        censys_args["api_secret"] = args.api_secret

    c = SearchClient(**censys_args)

    index: CensysSearchAPIv2 = getattr(c.v2, args.index_type)

    view_args = {}
    write_args = {
        "file_format": "json" if args.output else "screen",
        "file_path": args.output,
    }

    if args.at_time:
        view_args["at_time"] = args.at_time

    document = index.view(args.document_id, **view_args)

    try:
        write_file(document, **write_args)
    except ValueError as error:  # pragma: no cover
        console.print(f"Error writing log file. Error: {error}")


def include(parent_parser: argparse._SubParsersAction, parents: dict):
    """Include this subcommand into the parent parser.

    Args:
        parent_parser (argparse._SubParsersAction): Parent parser.
        parents (dict): Parent arg parsers.
    """
    view_parser = parent_parser.add_parser(
        "view",
        description="View a document in Censys Search by providing a document \
            id and the resource index",
        help="view document",
        parents=[parents["auth"]],
    )
    view_parser.add_argument(
        "document_id",
        type=str,
        help="a document id (IP address) to view",
    )
    view_parser.add_argument(
        "--index-type",
        type=str,
        default="hosts",
        choices=V2_INDEXES,
        metavar="|".join(V2_INDEXES),
        help="which resource index to query",
    )
    view_parser.add_argument(
        "--at-time",
        type=valid_datetime_type,
        metavar="YYYY-MM-DD (HH:mm)",
        help="Fetches a document at a given point in time",
    )
    view_parser.add_argument(
        "-o",
        "--output",
        type=str,
        help="json output file path",
    )
    view_parser.add_argument(
        "-O",
        "--open",
        action="store_true",
        help="open document in browser",
    )
    view_parser.set_defaults(func=cli_view)



"""Interact with the Censys Search Cert API."""
from typing import List, Optional, Tuple

from .api import CensysSearchAPIv2


class CensysCerts(CensysSearchAPIv2):
    """Interacts with the Certs index.

    Please note that this class represents only the v2 API endpoints. The v1 API endpoints (search, view, and report) are avilable only from CensysCertificates.

    Examples:
        Inits Censys Certs.

        >>> from censys.search import CensysCerts
        >>> c = CensysCerts()

        Search for hosts by sha256fp.

        >>> c.get_hosts_by_cert("fb444eb8e68437bae06232b9f5091bccff62a768ca09e92eb5c9c2cf9d17c426")
        (
            [
                {
                    "ip": "string",
                    "name": "string",
                    "observed_at": "2021-08-02T14:56:38.711Z",
                    "first_observed_at": "2021-08-02T14:56:38.711Z",
                }
            ],
            {
                "next": "nextCursorToken",
            },
        )
    """

    INDEX_NAME = "certificates"
    """Name of Censys Index."""

    def view(self, document_id: str):  # type: ignore # noqa: D102
        """This function acts as a placeholder for the upcoming Censys Search Cert v2 API.

        To view a certificate, please use `CensysCertificates`.

        Raises:
            NotImplementedError: CensysCerts.view is not implemented yet. Please use CensysCertificates.view instead.

        :meta private:
        """
        raise NotImplementedError(
            "CensysCerts.view is not implemented yet. Please use CensysCertificates.view instead."
        )

    def search(  # type: ignore # noqa: D102
        self, query: str, per_page: Optional[int], cursor: Optional[str], pages: int
    ):
        """This function acts as a placeholder for the upcoming Censys Search Cert v2 API.

        To search for certificates, please use `CensysCertificates`.

        Raises:
            NotImplementedError: CensysCerts.search is not implemented yet. Please use CensysCertificates.search instead.

        :meta private:
        """
        raise NotImplementedError(
            "CensysCerts.search is not implemented yet. Please use CensysCertificates.search instead."
        )

    def aggregate(self, query: str, field: str, num_buckets: Optional[int]):  # type: ignore # noqa: D102
        """This function acts as a placeholder for the upcoming Censys Search Cert v2 API.

        To aggregate/report certificates, please use `CensysCertificates`.

        Raises:
            NotImplementedError: CensysCerts.aggregate is not implemented yet. Please use CensysCertificates.report instead.

        :meta private:
        """
        raise NotImplementedError(
            "CensysCerts.aggregate is not implemented yet. Please use CensysCertificates.report instead."
        )

    def metadata(self):  # noqa: D102
        """This function acts as a placeholder for the upcoming Censys Search Cert v2 API.

        Raises:
            NotImplementedError: CensysCerts.metadata is not implemented yet.

        :meta private:
        """
        raise NotImplementedError("CensysCerts.metadata is not implemented yet.")

    def get_hosts_by_cert(
        self, sha256fp: str, cursor: Optional[str] = None
    ) -> Tuple[List[str], dict]:
        """Returns a list of hosts which contain services presenting this certificate.

        Args:
            sha256fp (str): The SHA-256 fingerprint of the requested certificate.
            cursor (str): Cursor token from the API response, which fetches the next page of hosts when added to the endpoint URL.

        Returns:
            Tuple[List[str], dict]: A list of hosts and a dictionary of the pagination cursors.
        """
        args = {"cursor": cursor}
        result = self._get(self.view_path + sha256fp + "/hosts", args)["result"]
        return result["hosts"], result["links"]

    def list_certs_with_tag(self, tag_id: str) -> List[str]:
        """Returns a list of certs which are tagged with the specified tag.

        Args:
            tag_id (str): The ID of the tag.

        Returns:
            List[str]: A list of certificate SHA 256 fingerprints.
        """
        certs = self._list_documents_with_tag(tag_id, "certificates", "certs")
        return [cert["fingerprint"] for cert in certs]



"""Interact with the Censys Clouds API."""
import datetime
from typing import Union

from .api import CensysAsmAPI

Since = Union[str, datetime.date, datetime.datetime]


def format_since_date(since: Since) -> str:
    """Formats since date as ISO 8601 format.

    Args:
        since (Since): Date.

    Returns:
        str: ISO 8601 formatted date string.
    """
    if isinstance(since, (datetime.date, datetime.datetime)):
        return since.strftime("%Y-%m-%d")
    return since


class Clouds(CensysAsmAPI):
    """Clouds API class."""

    base_path = "/v1/clouds"

    def get_host_counts(
        self,
        since: Since,
    ) -> dict:
        """Retrieve host counts by cloud.

        Hosts found after the date provided in the `since` parameter will be included in the new asset counts.

        Args:
            since (Since): Date to include hosts from.

        Returns:
            dict: Host count result.
        """
        since = format_since_date(since)
        return self._get(f"{self.base_path}/hostCounts/{since}")

    def get_domain_counts(self, since: Since) -> dict:
        """Retrieve domain counts by cloud.

        Args:
            since (Since): Date to include domains from.

        Returns:
            dict: Domain count result.
        """
        since = format_since_date(since)
        return self._get(f"{self.base_path}/domainCounts/{since}")

    def get_object_store_counts(self, since: Since) -> dict:
        """Retrieve object store counts by cloud.

        Args:
            since (Since): Date to include object stores from.

        Returns:
            dict: Object store count result.
        """
        since = format_since_date(since)
        return self._get(f"{self.base_path}/objectStoreCounts/{since}")

    def get_subdomain_counts(self, since: Since) -> dict:
        """Retrieve subdomain counts by cloud.

        Args:
            since (Since): Date to include subdomains from.

        Returns:
            dict: Subdomain count result.
        """
        since = format_since_date(since)
        return self._get(f"{self.base_path}/subdomainCounts/{since}")

    def get_unknown_counts(self) -> dict:
        """Retrieve known and unknown counts for hosts by cloud.

        Returns:
            dict: Unknown count result.
        """
        return self._get(f"{self.base_path}/unknownCounts")



"""An easy-to-use and lightweight API wrapper for Censys ASM (app.censys.io)."""
from .assets import Assets, CertificatesAssets, DomainsAssets, HostsAssets
from .client import AsmClient
from .clouds import Clouds
from .events import Events
from .risks import Risks
from .seeds import Seeds

__all__ = [
    "AsmClient",
    "Clouds",
    "Events",
    "Risks",
    "Seeds",
    "Assets",
    "CertificatesAssets",
    "DomainsAssets",
    "HostsAssets",
]



"""Interact with the Censys Subdomain Assets API."""
from typing import Any, Dict, Iterator, List, Optional

from .assets import Assets


class SubdomainsAssets(Assets):
    """Subdomains Assets API class."""

    def __init__(self, domain: str, *args, **kwargs):
        """Inits SubdomainsAssets.

        Args:
            domain: Name of the parent domain
            *args: Variable length argument list.
            **kwargs: Arbitrary keyword arguments.
        """
        # Rewrite the path because it chains off of the parent domain
        super().__init__(f"domains/{domain}/subdomains", *args, **kwargs)

    def get_assets(
        self,
        page_number: int = 1,
        page_size: Optional[int] = None,
        tag: Optional[List[str]] = None,
        tag_operator: Optional[str] = None,
        source: Optional[List[str]] = None,
        discovery_trail: Optional[bool] = None,
    ) -> Iterator[dict]:
        """Requests assets data.

        Override for subdomains due to it return value in a different key.

        Args:
            page_number (int): Optional; Page number to begin at when searching.
            page_size (int): Optional; Page size for retrieving assets.
            tag (list): Optional; List of tags to search for.
            tag_operator (str): Optional; Operator to use when searching for tags.
            source (list): Optional; List of sources to search for.
            discovery_trail (bool): Optional; Bool indicating whether to return discovery trail.

        Yields:
            dict: The assets result returned.
        """
        args: Dict[str, Any] = {}
        if tag:
            args["tag"] = tag
        if tag_operator:
            args["tagOperator"] = tag_operator
        if source:
            args["source"] = source
        if discovery_trail:
            args["discoveryTrail"] = discovery_trail
        yield from self._get_page(
            self.base_path,
            page_number=page_number,
            page_size=page_size,
            args=args,
            keyword="subdomains",
        )



"""Interact with the Censys Domain Assets API."""
from typing import Iterator, Optional

from .assets import Assets


class DomainsAssets(Assets):
    """Domains Assets API class."""

    def __init__(self, *args, **kwargs):
        """Inits DomainsAssets.

        Args:
            *args: Variable length argument list.
            **kwargs: Arbitrary keyword arguments.
        """
        super().__init__("domains", *args, **kwargs)

    def get_subdomains(
        self, domain: str, page_number: int = 1, page_size: Optional[int] = None
    ) -> Iterator[dict]:
        """List all subdomains of the parent domain.

        Args:
            domain: (str): Parent domain to query.
            page_number (int): Optional; Page number to begin at when searching.
            page_size (int): Optional; Page size for retrieving assets.

        Yields:
            dict: The assets result returned.
        """
        yield from self._get_page(
            f"{self.base_path}/{domain}/subdomains",
            page_number=page_number,
            page_size=page_size,
            keyword="subdomains",
        )



"""Interact with the Censys Certificate Assets API."""
from .assets import Assets


class CertificatesAssets(Assets):
    """Certificates Assets API class."""

    def __init__(self, *args, **kwargs):
        """Inits CertificatesAssets.

        Args:
            *args: Variable length argument list.
            **kwargs: Arbitrary keyword arguments.
        """
        super().__init__("certificates", *args, **kwargs)



"""Censys ASM CLI."""
import argparse
import json
import sys
from typing import Dict, List
from xml.etree import ElementTree

from rich.prompt import Confirm, Prompt

from censys.asm.seeds import SEED_TYPES, Seeds
from censys.cli.utils import console
from censys.common.config import DEFAULT, get_config, write_config
from censys.common.exceptions import CensysUnauthorizedException


def cli_asm_config(_: argparse.Namespace):  # pragma: no cover
    """Config asm subcommand.

    Args:
        _: Argparse Namespace.
    """
    api_key_prompt = "Censys ASM API Key"

    config = get_config()
    api_key = config.get(DEFAULT, "asm_api_key")

    if api_key:
        key_len = len(api_key) - 4
        redacted_api_key = api_key.replace(api_key[:key_len], key_len * "*")
        api_key_prompt = f"{api_key_prompt} [cyan]({redacted_api_key})[/cyan]"

    api_key = Prompt.ask(api_key_prompt, console=console) or api_key

    if not api_key:
        console.print("Please enter valid credentials")
        sys.exit(1)

    color = Confirm.ask(
        "Do you want color output?", default=True, show_default=False, console=console
    )
    config.set(DEFAULT, "color", "auto" if color else "")

    try:
        # Assumes that login was successfully
        config.set(DEFAULT, "asm_api_key", api_key)

        write_config(config)
        console.print("\nSuccessfully configured credentials")
        sys.exit(0)
    except CensysUnauthorizedException:
        console.print("Failed to authenticate")
        sys.exit(1)


def get_seeds_from_xml(file: str) -> List[Dict[str, str]]:
    """Get seeds from nmap xml.

    Args:
        file (str): Nmap xml file.

    Returns:
        List[Dict[str, str]]: List of seeds.
    """
    tree = ElementTree.parse(file)
    root = tree.getroot()
    ips = set()
    domains = set()
    for element in root.findall("host"):
        address_element = element.find("address")
        hostnames_element = element.find("hostnames")
        if address_element is not None and address_element.get("addrtype") == "ipv4":
            ip_address = address_element.get("addr")
            if ip_address:
                ips.add(ip_address)
        if hostnames_element is not None:
            hostname_elements = hostnames_element.findall("hostname")
            for hostname_element in hostname_elements:
                hostname_type = hostname_element.get("type")
                if hostname_type != "user":
                    continue
                hostname = hostname_element.get("name")
                if hostname:
                    domains.add(hostname)
    return [{"value": ip, "type": "IP_ADDRESS"} for ip in ips] + [
        {"value": domain, "type": "DOMAIN_NAME"} for domain in domains
    ]


def cli_add_seeds(args: argparse.Namespace):
    """Add seed subcommand.

    Args:
        args (Namespace): Argparse Namespace.
    """
    if args.input_file or args.json:
        if args.input_file:
            if args.input_file == "-":
                data = sys.stdin.read()
            else:
                with open(args.input_file) as f:
                    data = f.read()
        else:
            data = args.json
        try:
            seeds = json.loads(data)
        except json.decoder.JSONDecodeError as e:
            console.print(f"Invalid json {e}")
            sys.exit(1)
    elif args.nmap_xml:
        try:
            seeds = get_seeds_from_xml(args.nmap_xml)
        except ElementTree.ParseError as e:
            console.print(f"Invalid xml {e}")
            sys.exit(1)

    seeds_to_add = []
    for seed in seeds:
        if isinstance(seed, dict):
            if "type" not in seed:
                seed["type"] = args.default_type
        elif isinstance(seed, str):
            seed = {"value": seed, "type": args.default_type}
        else:
            console.print(f"Invalid seed {seed}")
            sys.exit(1)
        if "label" not in seed:
            seed["label"] = args.label_all
        seeds_to_add.append(seed)

    s = Seeds(args.api_key)
    to_add_count = len(seeds_to_add)
    res = s.add_seeds(seeds_to_add)
    added_seeds = res["addedSeeds"]
    added_count = len(added_seeds)
    if not added_count:
        console.print("No seeds were added. (Run with -v to get more info)")
        if not args.verbose:
            sys.exit(1)
    else:
        console.print(f"Added {added_count} seeds.")
    if added_count < to_add_count:
        console.print(f"Seeds not added: {to_add_count - added_count}")
        if args.verbose:  # pragma: no cover
            console.print(
                "The following seed(s) were not able to be added as they already exist or are reserved."
            )
            for seed in seeds_to_add:
                if not any([s for s in added_seeds if seed["value"] == s["value"]]):
                    console.print_json(seed)


def include(parent_parser: argparse._SubParsersAction, parents: dict):
    """Include this subcommand into the parent parser.

    Args:
        parent_parser (argparse._SubParsersAction): Parent parser.
        parents (dict): Parent arg parsers.
    """
    asm_parser = parent_parser.add_parser(
        "asm", description="Interact with the Censys ASM API", help="interact with ASM"
    )
    asm_subparser = asm_parser.add_subparsers()

    # Add config command
    config_parser = asm_subparser.add_parser(
        "config",
        description="Configure Censys ASM API Settings",
        help="configure Censys ASM API settings",
    )
    config_parser.set_defaults(func=cli_asm_config)

    # Add seed command
    add_parser = asm_subparser.add_parser(
        "add-seeds",
        description="Add seeds to ASM",
        help="add seeds",
        parents=[parents["asm_auth"]],
    )
    add_parser.add_argument(
        "--default-type",
        help="type of the seed(s) if type is not already provided (default: %(default)s)",
        choices=SEED_TYPES,
        default="IP_ADDRESS",
    )
    add_parser.add_argument(
        "--label-all",
        help='label to apply to all seeds (default: "")',
        type=str,
        default="",
    )
    add_parser.add_argument(
        "-v",
        "--verbose",
        help="verbose output",
        action="store_true",
    )
    seeds_group = add_parser.add_mutually_exclusive_group(required=True)
    seeds_group.add_argument(
        "--input-file",
        "-i",
        help="input file name containing valid json seeds (use - for stdin)",
        type=str,
    )
    seeds_group.add_argument(
        "--json", "-j", help="input string containing valid json seeds", type=str
    )
    seeds_group.add_argument(
        "--nmap-xml", help="input file name containing valid xml nmap output", type=str
    )
    add_parser.set_defaults(func=cli_add_seeds)



"""Censys CLI utilities."""
import argparse
import csv
import datetime
import json
import os.path
import sys
from typing import Any, Dict, List, Optional, Union

from rich.console import Console

from censys.common.config import DEFAULT, get_config
from censys.common.deprecation import DeprecationDecorator

Results = Union[List[dict], Dict[str, Any]]

V1_INDEXES = ["certs"]
V2_INDEXES = ["hosts"]
INDEXES = V1_INDEXES + V2_INDEXES

config = get_config()
color = config.get(DEFAULT, "color")
color_system = "auto" if color else None
console = Console(color_system=color_system)  # type: ignore
err_console = Console(color_system=color_system, file=sys.stderr)  # type: ignore


def print_wrote_file(file_path: str):
    """Print wrote file confirmation.

    Args:
        file_path (str): Name of the file to write to on the disk.
    """
    abs_file_path = os.path.abspath(file_path)
    console.print(f"Wrote results to file {abs_file_path}", soft_wrap=True)


@DeprecationDecorator("CSV output is deprecated and will be removed in the future.")
def _write_csv(file_path: str, search_results: Results, fields: List[str]):
    """Write search results to a new file in CSV format.

    Args:
        file_path (str): Name of the file to write to on the disk.
        search_results (Results): A list of results from the query.
        fields (List[str]): A list of fields to write as headers.
    """
    with open(file_path, "w") as output_file:
        if search_results and isinstance(search_results, list):
            # Get the header row from the first result
            writer = csv.DictWriter(output_file, fieldnames=fields)
            writer.writeheader()

            for result in search_results:
                # Use the Dict writer to process and write results to CSV
                writer.writerow(result)

    print_wrote_file(file_path)


def _write_json(file_path: str, search_results: Results):
    """Write search results to a new file in JSON format.

    Args:
        file_path (str): Name of the file to write to on the disk.
        search_results (Results): A list of results from the query.
    """
    with open(file_path, "w") as output_file:
        # Since the results are already in JSON, just write them to a file.
        json.dump(search_results, output_file, indent=4)

    print_wrote_file(file_path)


def _write_screen(search_results: Results):  # pragma: no cover
    """Writes search results to standard output.

    Args:
        search_results (Results): A list of results from the query.
    """
    config = get_config()
    if config.get(DEFAULT, "color"):
        console.print_json(data=search_results)
    else:
        print(json.dumps(search_results, indent=4))


def write_file(
    results_list: Results,
    file_format: Optional[str] = None,
    file_path: Optional[str] = None,
    csv_fields: Optional[List[str]] = None,
):
    """Maps formats and writes results.

    Args:
        results_list (Results): A list of results from the API query.
        file_format (str): Optional; The format of the output.
        file_path (str): Optional; A path to write results to.
        csv_fields (List[str]): Optional; A list of fields to write to CSV.
    """
    if file_format and isinstance(file_format, str):
        file_format = file_format.lower()

    if not file_path:
        file_path = "temp-out.json"

    if file_format == "json":
        _write_json(file_path, results_list)
    elif file_format == "csv":
        if csv_fields is None:  # pragma: no cover
            csv_fields = []
        _write_csv(file_path, results_list, fields=csv_fields)
    else:
        _write_screen(results_list)


def valid_datetime_type(datetime_str: str) -> datetime.datetime:
    """Custom argparse type for user datetime values from arg.

    Args:
        datetime_str (str): A string representing a datetime.

    Raises:
        ArgumentTypeError: If the datetime string is invalid.

    Returns:
        datetime.datetime: A datetime object.
    """
    try:
        return datetime.datetime.strptime(datetime_str, "%Y-%m-%d %H:%M")
    except ValueError:
        try:
            return datetime.datetime.strptime(datetime_str, "%Y-%m-%d")
        except ValueError:
            msg = f"Given datetime ({datetime_str}) is not valid! Expected format: 'YYYY-MM-DD' or 'YYYY-MM-DD HH:mm'."
            raise argparse.ArgumentTypeError(msg)



"""Base for interacting with the Censys ASM API."""
import os
from math import inf
from typing import Iterator, Optional, Type

from requests.models import Response

from censys.common.base import CensysAPIBase
from censys.common.config import DEFAULT, get_config
from censys.common.exceptions import (
    CensysAsmException,
    CensysException,
    CensysExceptionMapper,
)


class CensysAsmAPI(CensysAPIBase):
    """This is the base class for ASM's Seeds, Assets, and Events classes."""

    DEFAULT_URL: str = "https://app.censys.io/api"
    """Default ASM API base URL."""

    def __init__(self, api_key: Optional[str] = None, **kwargs):
        """Inits CensysAsmAPI.

        Args:
            api_key (str): Optional; The API Key provided by Censys.
            **kwargs: Arbitrary keyword arguments.

        Raises:
            CensysException: Base Exception Class for the Censys API.
        """
        url = kwargs.pop("url", self.DEFAULT_URL)
        CensysAPIBase.__init__(self, url=url, **kwargs)

        # Gets config file
        config = get_config()

        # Try to get credentials
        self._api_key = (
            api_key
            or os.getenv("CENSYS_ASM_API_KEY")
            or config.get(DEFAULT, "asm_api_key")
        )

        if not self._api_key:
            raise CensysException("No ASM API key configured.")

        self._session.headers.update(
            {"Content-Type": "application/json", "Censys-Api-Key": self._api_key}
        )

    def _get_exception_class(  # type: ignore
        self, res: Response
    ) -> Type[CensysAsmException]:
        return CensysExceptionMapper.ASM_EXCEPTIONS.get(
            res.json().get("errorCode"), CensysAsmException
        )

    def _get_page(
        self,
        path: str,
        page_number: int = 1,
        page_size: Optional[int] = None,
        args: Optional[dict] = None,
        keyword: str = "assets",
    ) -> Iterator[dict]:
        """Fetches paginated ASM resource API results.

        Args:
            path (str): The API url endpoint.
            page_number (int): Optional; Page number to begin at when getting results.
            page_size (int):
                Optional; Number of results to return per HTTP request. Defaults to 500.
            args (dict): Optional; URL args that are mapped to params.
            keyword (str): Optional; The keyword to iterate over in the results.

        Yields:
            dict: The resource result returned.
        """
        total_pages = inf
        args = args or {}

        while page_number <= total_pages:
            args.update({"pageNumber": page_number, "pageSize": page_size or 500})

            res = self._get(path, args=args)
            page_number = int(res["pageNumber"]) + 1
            total_pages = int(res["totalPages"])

            yield from res[keyword]

    def _get_logbook_page(
        self, path: str, args: Optional[dict] = None
    ) -> Iterator[dict]:
        """Fetches paginated ASM logbook API events.

        Args:
            path (str): The API url endpoint.
            args (dict): Optional; URL args that are mapped to params (cursor).

        Yields:
            dict: The event result returned.
        """
        end_of_events = False

        while not end_of_events:
            res = self._get(path, args=args)
            end_of_events = res["endOfEvents"]
            args = {"cursor": res["nextCursor"]}

            yield from res["events"]



"""Base for interacting with the Censys Search API."""
import os
from concurrent.futures import ThreadPoolExecutor, as_completed
from typing import Any, Dict, Iterable, Iterator, List, Optional, Type

from requests.models import Response

from censys.common.base import CensysAPIBase
from censys.common.config import DEFAULT, get_config
from censys.common.exceptions import (
    CensysException,
    CensysExceptionMapper,
    CensysSearchException,
)
from censys.common.types import Datetime
from censys.common.utils import format_rfc3339

INDEX_TO_KEY = {"hosts": "ip"}


class CensysSearchAPIv2(CensysAPIBase):
    """This class is the base class for the Hosts index.

    Examples:
        >>> c = CensysSearchAPIv2()
    """

    DEFAULT_URL: str = "https://search.censys.io/api"
    """Default Search API base URL."""
    INDEX_NAME: str = ""
    """Name of Censys Index."""

    def __init__(
        self, api_id: Optional[str] = None, api_secret: Optional[str] = None, **kwargs
    ):
        """Inits CensysSearchAPIv2.

        See CensysAPIBase for additional arguments.

        Args:
            api_id (str): Optional; The API ID provided by Censys.
            api_secret (str): Optional; The API secret provided by Censys.
            **kwargs: Arbitrary keyword arguments.

        Raises:
            CensysException: Base Exception Class for the Censys API.
        """
        CensysAPIBase.__init__(self, kwargs.pop("url", self.DEFAULT_URL), **kwargs)

        # Gets config file
        config = get_config()

        # Try to get credentials
        self._api_id = (
            api_id or os.getenv("CENSYS_API_ID") or config.get(DEFAULT, "api_id")
        )
        self._api_secret = (
            api_secret
            or os.getenv("CENSYS_API_SECRET")
            or config.get(DEFAULT, "api_secret")
        )
        if not self._api_id or not self._api_secret:
            raise CensysException("No API ID or API secret configured.")

        self._session.auth = (self._api_id, self._api_secret)

        # Generate concrete paths to be called
        self.view_path = f"/v2/{self.INDEX_NAME}/"
        self.search_path = f"/v2/{self.INDEX_NAME}/search"
        self.aggregate_path = f"/v2/{self.INDEX_NAME}/aggregate"
        self.metadata_path = f"/v2/metadata/{self.INDEX_NAME}"
        self.tags_path = "/v2/tags"
        self.account_path = "/v1/account"

    def _get_exception_class(  # type: ignore
        self, res: Response
    ) -> Type[CensysSearchException]:
        return CensysExceptionMapper.SEARCH_EXCEPTIONS.get(
            res.status_code, CensysSearchException
        )

    def account(self) -> dict:
        """Gets the current account's query quota.

        Returns:
            dict: Quota response.
        """
        return self._get(self.account_path)

    def quota(self) -> dict:
        """Returns metadata of a given search query.

        Returns:
            dict: The metadata of the result set returned.
        """
        return self.account()["quota"]

    class Query(Iterable):
        """Query class that is callable and iterable.

        Object Searches the given index for all records that match the given query.
        For more details, see our documentation: https://search.censys.io/api
        """

        def __init__(
            self,
            api: "CensysSearchAPIv2",
            query: str,
            per_page: Optional[int] = None,
            cursor: Optional[str] = None,
            pages: int = 1,
            **kwargs: Any,
        ):
            """Inits Query.

            Args:
                api (CensysSearchAPIv2): Parent API object.
                query (str): The query to be executed.
                per_page (int): Optional; The number of results to be returned for each page. Defaults to 100.
                cursor (int): Optional; The cursor of the desired result set.
                pages (int): Optional; The number of pages returned. Defaults to 1. If you set this to -1, it will return all pages.
                **kwargs (Any): Optional; Additional arguments to be passed to the query.
            """
            self.api = api
            self.query = query
            self.per_page = per_page
            self.cursor = cursor
            self.nextCursor: Optional[str] = None
            self.page = 1
            if pages == -1:
                self.pages = float("inf")
            else:
                self.pages = pages
            self.extra_args = kwargs

        def __call__(self, per_page: Optional[int] = None) -> List[dict]:
            """Search current index.

            Args:
                per_page (int): Optional; The number of results to be returned for each page. Defaults to 100.

            Raises:
                StopIteration: Raised when pages have been already received.

            Returns:
                List[dict]: One page worth of result hits.
            """
            if self.page > self.pages:
                raise StopIteration

            payload = self.api.raw_search(
                query=self.query,
                per_page=per_page or self.per_page or 100,
                cursor=self.nextCursor or self.cursor,
                **self.extra_args,
            )
            self.page += 1
            result = payload["result"]
            self.nextCursor = result["links"]["next"]
            if result["total"] == 0 or not self.nextCursor:
                self.pages = 0
            return result["hits"]

        def __next__(self) -> List[dict]:
            """Gets next page of search results.

            Returns:
                List[dict]: One page worth of result hits.
            """
            return self.__call__()

        def __iter__(self) -> Iterator[List[dict]]:
            """Gets Iterator.

            Returns:
                Iterable: Returns self.
            """
            return self

        def view_all(self, max_workers: int = 20) -> Dict[str, dict]:
            """View each document returned from query.

            Please note that each result returned by the query will be looked up using the view method.

            Args:
                max_workers (int): The number of workers to use. Defaults to 20.

            Returns:
                Dict[str, dict]: Dictionary mapping documents to that document's result set.
            """
            results = {}

            document_key = INDEX_TO_KEY.get(self.api.INDEX_NAME, "ip")

            with ThreadPoolExecutor(max_workers) as executor:
                threads = {}
                for hit in self.__call__():
                    hit_key = hit[document_key]
                    if "name" in hit:
                        hit_key += "+" + hit["name"]
                    threads[executor.submit(self.api.view, hit_key)] = hit_key

                for task in as_completed(threads):
                    document_id = threads[task]
                    try:
                        results[document_id] = task.result()
                    except Exception as e:
                        results[document_id] = {"error": str(e)}

            return results

    def search(
        self,
        query: str,
        per_page: Optional[int] = None,
        cursor: Optional[str] = None,
        pages: int = 1,
        **kwargs: Any,
    ) -> Query:
        """Search current index.

        Searches the given index for all records that match the given query.
        For more details, see our documentation: https://search.censys.io/api

        Args:
            query (str): The query to be executed.
            per_page (int): Optional; The number of results to be returned for each page. Defaults to 100.
            cursor (int): Optional; The cursor of the desired result set.
            pages (int): Optional; The number of pages returned. Defaults to 1.
            **kwargs (Any): Optional; Additional arguments to be passed to the query.

        Returns:
            Query: Query object that can be a callable or an iterable.
        """
        return self.Query(self, query, per_page, cursor, pages, **kwargs)

    def raw_search(
        self,
        query: str,
        per_page: Optional[int] = None,
        cursor: Optional[str] = None,
        **kwargs: Any,
    ) -> dict:
        """Search current index.

        Searches the given index for all records that match the given query.
        This method does no automatic pagination or post processing.

        Args:
            query (str): The query to be executed.
            per_page (int): Optional; The number of results to be returned for each page. Defaults to 100.
            cursor (int): Optional; The cursor of the desired result set.
            **kwargs (Any): Optional; Additional arguments to be passed to the query.

        Returns:
            dict: The raw result set.
        """
        args = {
            "q": query,
            "per_page": per_page or 100,
            "cursor": cursor,
            **kwargs,
        }
        return self._get(self.search_path, args)

    def view(
        self,
        document_id: str,
        at_time: Optional[Datetime] = None,
    ) -> dict:
        """View document from current index.

        View the current structured data we have on a specific document.
        For more details, see our documentation: https://search.censys.io/api

        Args:
            document_id (str): The ID of the document you are requesting.
            at_time ([str, datetime.date, datetime.datetime]):
                Optional; Fetches a document at a given point in time.

        Returns:
            dict: The result set returned.
        """
        args = {}
        if at_time:
            args["at_time"] = format_rfc3339(at_time)

        return self._get(self.view_path + document_id, args)["result"]

    def bulk_view(
        self,
        document_ids: List[str],
        at_time: Optional[Datetime] = None,
        max_workers: int = 20,
    ) -> Dict[str, dict]:
        """Bulk view documents from current index.

        View the current structured data we have on a list of documents.
        For more details, see our documentation: https://search.censys.io/api

        Args:
            document_ids (List[str]): The IDs of the documents you are requesting.
            at_time ([str, datetime.date, datetime.datetime]):
                Optional; Fetches a document at a given point in time.
            max_workers (int): The number of workers to use. Defaults to 20.

        Returns:
            Dict[str, dict]: Dictionary mapping document IDs to that document's result set.
        """
        if at_time:
            at_time = format_rfc3339(at_time)

        documents = {}
        with ThreadPoolExecutor(max_workers) as executor:
            threads = {
                executor.submit(self.view, document_id, at_time): document_id
                for document_id in document_ids
            }

            for task in as_completed(threads):
                document_id = threads[task]
                try:
                    documents[document_id] = task.result()
                except Exception as e:
                    documents[document_id] = {"error": str(e)}

        return documents

    def aggregate(
        self, query: str, field: str, num_buckets: Optional[int] = None, **kwargs: Any
    ) -> dict:
        """Aggregate current index.

        Creates a report on the breakdown of the values of a field in a result set.
        For more details, see our documentation: https://search.censys.io/api

        Args:
            query (str): The query to be executed.
            field (str): The field you are running a breakdown on.
            num_buckets (int): Optional; The maximum number of values. Defaults to 50.
            **kwargs (Any): Optional; Additional arguments to be passed to the query.

        Returns:
            dict: The result set returned.
        """
        args = {"q": query, "field": field, "num_buckets": num_buckets, **kwargs}
        return self._get(self.aggregate_path, args)["result"]

    def metadata(self) -> dict:
        """Get current index metadata.

        Returns:
            dict: The result set returned.
        """
        return self._get(self.metadata_path)["result"]

    # Comments

    def get_comments(self, document_id: str) -> List[dict]:
        """Get comments for a document.

        Args:
            document_id (str): The ID of the document you are requesting.

        Returns:
            List[dict]: The list of comments.
        """
        return self._get(self.view_path + document_id + "/comments")["result"][
            "comments"
        ]

    def add_comment(self, document_id: str, contents: str) -> dict:
        """Add comment to a document.

        Args:
            document_id (str): The ID of the document you are requesting.
            contents (str): The contents of the comment.

        Returns:
            dict: The result set returned.
        """
        return self._post(
            self.view_path + document_id + "/comments", data={"contents": contents}
        )["result"]

    def delete_comment(self, document_id: str, comment_id: str) -> dict:
        """Delete comment from a document.

        Args:
            document_id (str): The ID of the document you are requesting.
            comment_id (str): The ID of the comment you are requesting.

        Returns:
            dict: The result set returned.
        """
        return self._delete(self.view_path + document_id + "/comments/" + comment_id)

    def update_comment(self, document_id: str, comment_id: str, contents: str) -> dict:
        """Update comment from a document.

        Args:
            document_id (str): The ID of the document you are requesting.
            comment_id (str): The ID of the comment you are requesting.
            contents (str): The contents of the comment.

        Returns:
            dict: The result set returned.
        """
        return self._put(
            self.view_path + document_id + "/comments/" + comment_id,
            data={"contents": contents},
        )

    # Tags

    def list_all_tags(self) -> List[dict]:
        """List all tags.

        Returns:
            List[dict]: The list of tags.
        """
        return self._get(self.tags_path)["result"]["tags"]

    def create_tag(self, name: str, color: Optional[str] = None) -> dict:
        """Create a tag.

        Args:
            name (str): The name of the tag.
            color (str): Optional; The color of the tag.

        Returns:
            dict: The result set returned.
        """
        tag_def: Dict[str, Any] = {"name": name}
        if color:
            tag_def["metadata"] = {"color": color}
        return self._post(self.tags_path, data=tag_def)["result"]

    def get_tag(self, tag_id: str) -> dict:
        """Get a tag.

        Args:
            tag_id (str): The ID of the tag.

        Returns:
            dict: The result set returned.
        """
        return self._get(self.tags_path + "/" + tag_id)["result"]

    def update_tag(self, tag_id: str, name: str, color: Optional[str] = None) -> dict:
        """Update a tag.

        Args:
            tag_id (str): The ID of the tag.
            name (str): The name of the tag.
            color (str): The color of the tag.

        Returns:
            dict: The result set returned.
        """
        tag_def: Dict[str, Any] = {"name": name}
        if color:
            tag_def["metadata"] = {"color": color}
        return self._put(
            self.tags_path + "/" + tag_id,
            data=tag_def,
        )["result"]

    def delete_tag(self, tag_id: str):
        """Delete a tag.

        Args:
            tag_id (str): The ID of the tag.
        """
        self._delete(self.tags_path + "/" + tag_id)

    def _list_documents_with_tag(
        self, tag_id: str, endpoint: str, keyword: str
    ) -> List[dict]:
        """List documents by tag.

        Args:
            tag_id (str): The ID of the tag.
            endpoint (str): The endpoint to be called.
            keyword (str): The keyword to be used in the endpoint.

        Returns:
            List[dict]: The list of documents.
        """
        return self._get(self.tags_path + "/" + tag_id + "/" + endpoint)["result"][
            keyword
        ]

    def list_tags_on_document(self, document_id: str) -> List[dict]:
        """List tags on a document.

        Args:
            document_id (str): The ID of the document.

        Returns:
            List[dict]: The list of tags.
        """
        return self._get(self.view_path + document_id + "/tags")["result"]["tags"]

    def add_tag_to_document(self, document_id: str, tag_id: str):
        """Add a tag to a document.

        Args:
            document_id (str): The ID of the document.
            tag_id (str): The ID of the tag.
        """
        self._put(self.view_path + document_id + "/tags/" + tag_id)

    def remove_tag_from_document(self, document_id: str, tag_id: str):
        """Remove a tag from a document.

        Args:
            document_id (str): The ID of the document.
            tag_id (str): The ID of the tag.
        """
        self._delete(self.view_path + document_id + "/tags/" + tag_id)



"""Censys Version."""
try:  # pragma: no cover
    import importlib_metadata
except ImportError:  # pragma: no cover
    import importlib.metadata as importlib_metadata  # type: ignore

__version__: str = importlib_metadata.version("censys")



"""Interact with all Search APIs."""
from ..common.deprecation import DeprecationDecorator
from .v1 import CensysCertificates, CensysData
from .v2 import CensysCerts, CensysHosts


class SearchClient:
    """Client for interacting with all Search APIs.

    All indexes are passed the args and kwargs that are provided.

    Examples:
        Inits SearchClient.

        >>> from censys.search import SearchClient
        >>> c = SearchClient()

        Access both v1 and v2 indexes.

        >>> certs = c.v1.certificates # CensysCertificates()
        >>> data = c.v1.data # CensysData()
        >>> hosts = c.v2.hosts # CensysHosts()
        >>> certs = c.v2.certs # CensysCerts()
    """

    class _V1:
        """Class for v1 Search APIs."""

        @DeprecationDecorator(
            "The v1 Search API is deprecated and will be removed in the future."
        )
        def __init__(self, *args, **kwargs):
            """Inits V1.

            Args:
                *args: Variable length argument list.
                **kwargs: Arbitrary keyword arguments.
            """
            self.certificates = CensysCertificates(*args, **kwargs)
            # Alias certs to certificates
            self.certs = self.certificates
            self.data = CensysData(*args, **kwargs)

    class _V2:
        """Class for v2 Search APIs."""

        def __init__(self, *args, **kwargs):
            """Inits V2.

            Args:
                *args: Variable length argument list.
                **kwargs: Arbitrary keyword arguments.
            """
            self.hosts = CensysHosts(*args, **kwargs)
            self.certs = CensysCerts(*args, **kwargs)

    def __init__(self, *args, **kwargs):
        """Inits SearchClient.

        Args:
            *args: Variable length argument list.
            **kwargs: Arbitrary keyword arguments.
        """
        # Backwards compatability
        if len(args) == 2:
            kwargs["api_id"] = args[0]
            kwargs["api_secret"] = args[1]

        self.v1 = self._V1(**kwargs)
        self.v2 = self._V2(**kwargs)



"""Common utilities for the Censys Python SDK."""

import datetime

from .types import Datetime


def format_rfc3339(time: Datetime) -> str:
    """Formats a datetime object into an RFC3339 string.

    Args:
        time (Datetime): Datetime object to format.

    Returns:
        str: RFC3339 formatted string.
    """
    if isinstance(time, (datetime.date, datetime.datetime)):
        time = time.strftime("%Y-%m-%dT%H:%M:%S.%fZ")
    return time



"""Base for interacting with the Censys APIs."""
import json
import os
import warnings
from functools import wraps
from typing import Any, Callable, Optional, Type

import backoff
import requests
from requests.models import Response

from .exceptions import (
    CensysAPIException,
    CensysException,
    CensysJSONDecodeException,
    CensysTooManyRequestsException,
)
from .version import __version__


# Wrapper to make max_retries configurable at runtime
def _backoff_wrapper(method: Callable):
    @wraps(method)
    def _wrapper(self, *args, **kwargs):
        @backoff.on_exception(
            backoff.expo,
            (
                CensysTooManyRequestsException,
                requests.exceptions.Timeout,
            ),
            max_tries=self.max_retries,
            max_time=self.timeout,
        )
        def _impl():
            return method(self, *args, **kwargs)

        return _impl()

    return _wrapper


class CensysAPIBase:
    """This is the base class for API queries."""

    DEFAULT_TIMEOUT: int = 30
    """Default API timeout."""
    DEFAULT_USER_AGENT: str = f"censys/{__version__}"
    """Default API user agent."""
    DEFAULT_MAX_RETRIES: int = 5
    """Default max number of API retries."""

    def __init__(
        self,
        url: Optional[str] = None,
        timeout: Optional[int] = DEFAULT_TIMEOUT,
        max_retries: Optional[int] = DEFAULT_MAX_RETRIES,
        user_agent: Optional[str] = DEFAULT_USER_AGENT,
        proxies: Optional[dict] = None,
        cookies: Optional[dict] = None,
        **kwargs,
    ):
        """Inits CensysAPIBase.

        Args:
            url (str): Optional; The URL to make API requests.
            timeout (int): Optional; Timeout for API requests in seconds.
            max_retries (int):
                Optional; Max number of times to retry failed API requests.
            user_agent (str): Optional; Override User-Agent string.
            proxies (dict): Optional; Configure HTTP proxies.
            cookies (dict): Optional; Configure cookies.
            **kwargs: Arbitrary keyword arguments.

        Raises:
            CensysException: Base Exception Class for the Censys API.
        """
        # Get common request settings
        self.timeout = timeout
        self.max_retries = max_retries
        self._api_url = url or os.getenv("CENSYS_API_URL")

        if not self._api_url:
            raise CensysException("No API url configured.")

        # Create a session and set credentials
        self._session = requests.Session()
        if proxies:
            if "http" in proxies:
                warnings.warn("HTTP proxies will not be used.")
                proxies.pop("http", None)
            self._session.proxies.update(proxies)
        if cookies:
            self._session.cookies.update(cookies)
        self._session.headers.update(
            {
                "accept": "application/json, */8",
                "User-Agent": " ".join(
                    [
                        requests.utils.default_user_agent(),
                        user_agent
                        or kwargs.get("user_agent_identifier")
                        or self.DEFAULT_USER_AGENT,
                    ]
                ),
            }
        )

    @staticmethod
    def _get_exception_class(_: Response) -> Type[CensysAPIException]:
        """Maps HTTP status code or ASM error code to exception.

        Must be implemented by child class.

        Args:
            _ (Response): HTTP requests response object.

        Returns:
            Type[CensysAPIException]: Exception to raise.
        """
        return CensysAPIException

    @_backoff_wrapper
    def _make_call(
        self,
        method: Callable[..., Response],
        endpoint: str,
        args: Optional[dict] = None,
        data: Optional[Any] = None,
        **kwargs,
    ) -> dict:
        """Make API call.

        Wrapper functions for all our REST API calls checking for errors
        and decoding the responses.

        Args:
            method (Callable): Method to send HTTP request.
            endpoint (str): The path of API endpoint.
            args (dict): Optional; URL args that are mapped to params.
            data (Any): Optional; JSON data to serialize with request.
            **kwargs: Arbitrary keyword arguments to pass to method.

        Raises:
            censys_exception: Exception Class for the Censys API.
            CensysJSONDecodeException: Exception for decoding JSON.

        Returns:
            dict: Results from an API request.
        """
        if endpoint.startswith("/"):
            url = f"{self._api_url}{endpoint}"
        else:
            url = f"{self._api_url}/{endpoint}"

        request_kwargs = {
            "params": args or {},
            "timeout": self.timeout,
            **kwargs,
        }

        if data:
            request_kwargs["json"] = data

        res = method(url, **request_kwargs)

        if res.ok:
            # Check for a returned json body
            try:
                json_data = res.json()
                if "error" not in json_data:
                    return json_data
            # Successful request returned no json body in response
            except ValueError:
                return {
                    "code": res.status_code,
                    "status": res.reason,
                }

        try:
            json_data = res.json()
            message = json_data.get("error") or json_data.get("message")
            const = json_data.get("error_type") or json_data.get("status") or res.reason
            error_code = json_data.get("errorCode", "unknown")
            details = json_data.get("details", "unknown")
        except (ValueError, json.decoder.JSONDecodeError) as error:
            raise CensysJSONDecodeException(
                status_code=res.status_code,
                message=f"Response from {res.url} is not valid JSON and cannot be decoded.",
                body=res.text,
                const="badjson",
            ) from error

        censys_exception = self._get_exception_class(res)
        raise censys_exception(
            status_code=res.status_code,
            body=res.text,
            const=const,
            message=message,
            error_code=error_code,
            details=details,
        )

    def _get(self, endpoint: str, args: Optional[dict] = None, **kwargs) -> dict:
        return self._make_call(self._session.get, endpoint, args, **kwargs)

    def _post(
        self,
        endpoint: str,
        args: Optional[dict] = None,
        data: Optional[dict] = None,
        **kwargs,
    ) -> dict:
        return self._make_call(self._session.post, endpoint, args, data, **kwargs)

    def _put(
        self,
        endpoint: str,
        args: Optional[dict] = None,
        data: Optional[dict] = None,
        **kwargs,
    ) -> dict:
        return self._make_call(self._session.put, endpoint, args, data, **kwargs)

    def _patch(
        self,
        endpoint: str,
        args: Optional[dict] = None,
        data: Optional[dict] = None,
        **kwargs,
    ) -> dict:
        return self._make_call(self._session.patch, endpoint, args, data, **kwargs)

    def _delete(self, endpoint: str, args: Optional[dict] = None, **kwargs) -> dict:
        return self._make_call(self._session.delete, endpoint, args, **kwargs)



"""Interact with the Censys Search Data API."""
from .api import CensysSearchAPIv1


class CensysData(CensysSearchAPIv1):
    """Interacts with the Data index.

    For more details, see our documentation: https://search.censys.io/api
    """

    _PREFIX = "/data"
    """Endpoint prefix."""

    def get_series(self) -> dict:
        """Get data on the types of scans we regularly perform (series).

        Returns:
            dict: The result set returned.
        """
        return self._get(self._PREFIX)

    def view_series(self, series_id: str) -> dict:
        """Get data on a specific series.

        Args:
            series_id (str): The ID of the series.

        Returns:
            dict: The result set returned.
        """
        path = f"{self._PREFIX}/{series_id}"
        return self._get(path)

    def view_result(self, series_id: str, result_id: str) -> dict:
        """View a specific result of a specific series.

        Args:
            series_id (str): The ID of the series.
            result_id (str): The ID of the result.

        Returns:
            dict: The result set returned.
        """
        path = f"{self._PREFIX}/{series_id}/{result_id}"
        return self._get(path)



"""Base for interacting with the Censys Assets API."""
import re
from typing import Any, Dict, Iterator, List, Optional

from ..api import CensysAsmAPI
from censys.common.exceptions import CensysInvalidColorException

HEX_REGEX = re.compile(r"^#(?:[0-9a-fA-F]{3}){1,2}$")


class Assets(CensysAsmAPI):
    """Assets API class."""

    def __init__(self, asset_type: str, *args, **kwargs):
        """Inits Assets.

        Args:
            asset_type (str): Type of asset to interact with.
            *args: Variable length argument list.
            **kwargs: Arbitrary keyword arguments.
        """
        CensysAsmAPI.__init__(self, *args, **kwargs)
        self.base_path = f"/v1/assets/{asset_type}"

    def get_assets(
        self,
        page_number: int = 1,
        page_size: Optional[int] = None,
        tag: Optional[List[str]] = None,
        tag_operator: Optional[str] = None,
        source: Optional[List[str]] = None,
        discovery_trail: Optional[bool] = None,
    ) -> Iterator[dict]:
        """Requests assets data.

        Args:
            page_number (int): Optional; Page number to begin at when searching.
            page_size (int): Optional; Page size for retrieving assets.
            tag (list): Optional; List of tags to search for.
            tag_operator (str): Optional; Operator to use when searching for tags.
            source (list): Optional; List of sources to search for.
            discovery_trail (bool): Optional; Bool indicating whether to return discovery trail.

        Yields:
            dict: The assets result returned.
        """
        args: Dict[str, Any] = {}
        if tag:
            args["tag"] = tag
        if tag_operator:
            args["tagOperator"] = tag_operator
        if source:
            args["source"] = source
        if discovery_trail:
            args["discoveryTrail"] = discovery_trail
        yield from self._get_page(
            self.base_path, page_number=page_number, page_size=page_size, args=args
        )

    def get_asset_by_id(self, asset_id: str) -> dict:
        """Requests asset data by ID.

        Args:
            asset_id (str): Requested asset ID.

        Returns:
            dict: Asset search result.
        """
        path = f"{self.base_path}/{asset_id}"

        return self._get(path)

    def get_comments(
        self,
        asset_id: str,
        page_number: int = 1,
        page_size: Optional[int] = None,
    ) -> Iterator[dict]:
        """Requests comments on a specified asset.

        Args:
            asset_id (str): Asset ID for requested comments.
            page_number (int): Optional; Page number to begin at when searching.
            page_size (int): Optional; Page size for retrieving comments.

        Returns:
            generator: Comment search results.
        """
        path = f"{self.base_path}/{asset_id}/comments"

        return self._get_page(
            path, page_number=page_number, page_size=page_size, keyword="comments"
        )

    def get_comment_by_id(self, asset_id: str, comment_id: int) -> dict:
        """Requests a comment on a specified asset by comment ID.

        Args:
            asset_id (str): Asset ID for requested comments.
            comment_id (int): Requested comment ID.

        Returns:
            dict: Comment search result.
        """
        path = f"{self.base_path}/{asset_id}/comments/{comment_id}"

        return self._get(path)

    def add_comment(self, asset_id: str, comment: str) -> dict:
        """Adds a comment to a specified asset on the ASM platform.

        Args:
            asset_id (str): Asset ID to add comment to.
            comment (str): New comment text.

        Returns:
            dict: Added comment results.
        """
        path = f"{self.base_path}/{asset_id}/comments"
        data = {"markdown": str(comment)}

        return self._post(path, data=data)

    def delete_comment(self, asset_id: str, comment_id: int) -> dict:
        """Deletes a comment from a specified asset on the ASM platform by comment ID.

        Args:
            asset_id (str): Asset ID to delete comment from.
            comment_id (int): Comment ID to delete.

        Returns:
            dict: Deleted comment results.
        """
        path = f"{self.base_path}/{asset_id}/comments/{comment_id}"

        return self._delete(path)

    def add_tag(self, asset_id: str, name: str, color: Optional[str] = None) -> dict:
        """Adds a tag to a specified asset on the ASM platform.

        Args:
            asset_id (str): Asset ID to add tag to.
            name (str): New tag name.
            color (str): Optional; New tag color (hex).

        Returns:
            dict: Added tag results.
        """
        path = f"{self.base_path}/{asset_id}/tags"
        data = format_tag(name, color)

        return self._post(path, data=data)

    def delete_tag(self, asset_id: str, name: str) -> dict:
        """Deletes a tag from a specified asset on the ASM platform by tag name.

        Args:
            asset_id (str): Asset ID to delete tag from.
            name (str): Tag name to delete.

        Returns:
            dict: Deleted tag results.
        """
        path = f"{self.base_path}/{asset_id}/tags/{name}"

        return self._delete(path)


def format_tag(name: str, color: Optional[str] = None) -> dict:
    """Formats tag name and color request data.

    Args:
        name (str): Tag name.
        color (str): Optional; Tag color (hex).

    Raises:
        CensysInvalidColorException: Raised if color is not a valid hex color.

    Returns:
        dict: Formatted tag request data.
    """
    if color:
        if not HEX_REGEX.match(color):
            raise CensysInvalidColorException(10037, f"{color} is not a valid color.")
        return {"name": str(name), "color": str(color)}

    return {"name": str(name)}



"""Interact with the Censys Logbook API."""
import datetime
from typing import Iterator, List, Optional, Union

from .api import CensysAsmAPI


class Events(CensysAsmAPI):
    """Events API class."""

    base_path = "/v1/logbook"

    def get_cursor(
        self,
        start: Optional[Union[datetime.datetime, int]] = None,
        filters: Optional[List[str]] = None,
    ) -> str:
        """Requests a logbook cursor.

        Args:
            start ([datetime.datetime, int]): Optional; Timestamp or event ID to begin searching.
            filters (list): Optional; List of filters applied to logbook search results.

        Returns:
            str: Cursor result.
        """
        path = f"{self.base_path}-cursor"
        data = format_data(start=start, filters=filters)

        return self._post(path, data=data)["cursor"]

    def get_events(self, cursor: Optional[str] = None) -> Iterator[dict]:
        """Requests logbook events from inception or from the provided cursor.

        Args:
            cursor (str): Optional; Logbook cursor.

        Yields:
            dict: Logbook event.
        """
        args = {"cursor": cursor}

        yield from self._get_logbook_page(self.base_path, args)


class Filters:
    """Logbook filters class."""

    CERT = "CERT"
    CERT_RISK = "CERT_RISK"
    DOMAIN = "DOMAIN"
    DOMAIN_EXPIRATION_DATE = "DOMAIN_EXPIRATION_DATE"
    DOMAIN_MAIL_EXCHANGE_SERVER = "DOMAIN_MAIL_EXCHANGE_SERVER"
    DOMAIN_NAME_SERVER = "DOMAIN_NAME_SERVER"
    DOMAIN_REGISTRAR = "DOMAIN_REGISTRAR"
    DOMAIN_RISK = "DOMAIN_RISK"
    DOMAIN_SUBDOMAIN = "DOMAIN_SUBDOMAIN"
    HOST = "HOST"
    HOST_CERT = "HOST_CERT"
    HOST_PORT = "HOST_PORT"
    HOST_PROTOCOL = "HOST_PROTOCOL"
    HOST_RISK = "HOST_RISK"
    HOST_SOFTWARE = "HOST_SOFTWARE"
    HOST_VULNERABILITY = "HOST_VULNERABILITY"


def format_data(
    start: Optional[Union[datetime.datetime, int]] = None,
    filters: Optional[List[str]] = None,
) -> dict:
    """Formats cursor request data into a start date/id and filter list.

    Args:
        start ([datetime.datetime, int]): Optional; Timestamp or event ID to begin searching.
        filters (list): Optional; List of filters applied to logbook search results.

    Returns:
        dict: Formatted logbook cursor request data
    """
    data: dict = {}

    if filters:
        data["filter"] = {"type": filters}
    if isinstance(start, int):
        data["idFrom"] = start
    elif start:
        data["dateFrom"] = start

    return data



"""Censys config CLI."""
import argparse
import sys

from rich import box
from rich.table import Table

from censys.cli.utils import console
from censys.common.exceptions import CensysUnauthorizedException
from censys.search.v2.api import CensysSearchAPIv2


def cli_account(args: argparse.Namespace):  # pragma: no cover
    """Account subcommand.

    Args:
        args: Argparse Namespace.
    """
    try:
        client = CensysSearchAPIv2(args.api_id, args.api_secret)
        account = client.account()
        if args.json:
            console.print_json(data=account)
        else:
            table = Table(
                "Key", "Value", show_header=False, box=box.SQUARE, highlight=True
            )
            table.add_row("Email", account["email"])
            table.add_row("Login ID", account["login"])
            table.add_row("First Login", account["first_login"])
            table.add_row("Last Login", account["last_login"][:-7])
            quota = account["quota"]
            table.add_row(
                "Query Quota",
                f"{quota['used']} / {quota['allowance']} ({quota['used']/quota['allowance'] * 100 :.2f}%)",
            )
            table.add_row("Quota Resets At", quota["resets_at"])
            console.print(table)
        sys.exit(0)
    except CensysUnauthorizedException:
        console.print("Failed to authenticate")
        sys.exit(1)


def include(parent_parser: argparse._SubParsersAction, parents: dict):
    """Include this subcommand into the parent parser.

    Args:
        parent_parser (argparse._SubParsersAction): Parent parser.
        parents (dict): Parent arg parsers.
    """
    account_parser = parent_parser.add_parser(
        "account",
        description="Check Censys account details and quota",
        help="check Censys account details and quota",
        parents=[parents["auth"]],
    )
    account_parser.add_argument(
        "-j", "--json", action="store_true", help="Output in JSON format"
    )
    account_parser.set_defaults(func=cli_account)



"""Interact with the Censys Risks v2 API."""
from typing import List, Optional

from ..api import CensysAsmAPI


class Risksv2(CensysAsmAPI):
    """Risks v2 API class."""

    base_path = "/v2/risk"
    risk_instances_path = f"{base_path}-instances"
    risk_types_path = f"{base_path}-types"

    def get_risk_instances(
        self, include_events: Optional[bool] = None, accept: Optional[str] = None
    ) -> dict:
        """Retrieve risk instances.

        Args:
            include_events (bool): Optional; Whether to include events.
            accept (str): Optional; Accept header.

        Returns:
            dict: Risk instances result.
        """
        args = {"includeEvents": include_events}
        return self._get(
            self.risk_instances_path,
            args=args,
            headers={"Accept": accept} if accept else None,
        )

    def patch_risk_instances(self, data: dict) -> dict:
        """Patch risk instances.

        Args:
            data (dict): Risk instances data.

        Returns:
            dict: Risk instances result.
        """
        return self._patch(self.risk_instances_path, data=data)

    def search_risk_instances(self, data: dict, accept: Optional[str] = None) -> dict:
        """Search risk instances.

        Args:
            data (dict): Query data.
            accept (str): Optional; Accept header.

        Returns:
            dict: Risk instances result.
        """
        return self._post(
            f"{self.risk_instances_path}/search",
            data=data,
            headers={"Accept": accept} if accept else None,
        )

    def get_risk_instance(
        self, risk_instance_id: int, include_events: Optional[bool] = None
    ) -> dict:
        """Retrieve a risk instance.

        Args:
            risk_instance_id (int): Risk instance ID.
            include_events (bool): Optional; Whether to include events.

        Returns:
            dict: Risk instance result.
        """
        args = {"includeEvents": include_events}
        return self._get(f"{self.risk_instances_path}/{risk_instance_id}", args=args)

    def patch_risk_instance(self, risk_instance_id: int, data: dict) -> dict:
        """Patch a risk instance.

        Args:
            risk_instance_id (int): Risk instance ID.
            data (dict): Risk instance data.

        Returns:
            dict: Risk instance result.
        """
        return self._patch(f"{self.risk_instances_path}/{risk_instance_id}", data=data)

    def get_risk_types(
        self,
        sort: Optional[List[str]] = None,
        include_events: Optional[bool] = None,
        accept: Optional[str] = None,
    ) -> dict:
        """Retrieve risk types.

        Args:
            sort (list): Optional; Sort by field(s).
            include_events (bool): Optional; Whether to include events.
            accept (str): Optional; Accept header.

        Returns:
            dict: Risk types result.
        """
        args = {"sort": sort, "includeEvents": include_events}
        return self._get(
            self.risk_types_path,
            args=args,
            headers={"Accept": accept} if accept else None,
        )

    def get_risk_type(
        self, risk_type: str, include_events: Optional[bool] = None
    ) -> dict:
        """Retrieve a risk type.

        Args:
            risk_type (str): Risk type.
            include_events (bool): Optional; Whether to include events.

        Returns:
            dict: Risk type result.
        """
        args = {"includeEvents": include_events}
        return self._get(f"{self.risk_types_path}/{risk_type}", args=args)

    def patch_risk_type(self, risk_type: str, data: dict) -> dict:
        """Patch a risk type.

        Args:
            risk_type (str): Risk type.
            data (dict): Risk type data.

        Returns:
            dict: Risk type result.
        """
        return self._patch(f"{self.risk_types_path}/{risk_type}", data=data)



"""Censys subdomains CLI."""
import argparse
import json
import sys
from typing import List, Set

from censys.cli.utils import console, err_console
from censys.common.exceptions import (
    CensysException,
    CensysRateLimitExceededException,
    CensysUnauthorizedException,
)
from censys.search import CensysCertificates


def print_subdomains(subdomains: Set[str], as_json: bool = False):
    """Print subdomains.

    Args:
        subdomains (Set[str]): Subdomains.
        as_json (bool): Output in JSON format.
    """
    if as_json:
        console.print_json(json.dumps(list(subdomains)))
    else:
        for subdomain in subdomains:
            console.print(f"  - {subdomain}")


def cli_subdomains(args: argparse.Namespace):  # pragma: no cover
    """Subdomain subcommand.

    Args:
        args: Argparse Namespace.
    """
    subdomains = set()
    try:
        client = CensysCertificates(api_id=args.api_id, api_secret=args.api_secret)
        certificate_query = f"parsed.names: {args.domain}"

        with err_console.status(f"Querying {args.domain} subdomains"):
            certificates_search_results = client.search(
                certificate_query, fields=["parsed.names"], max_records=args.max_records
            )

            # Flatten the result, and remove duplicates
            for search_result in certificates_search_results:
                new_subdomains: List[str] = search_result.get("parsed.names", [])
                subdomains.update(
                    [
                        subdomain
                        for subdomain in new_subdomains
                        if subdomain.endswith(args.domain)
                    ]
                )

        # Don't make console prints if we're in json mode
        if not args.json:
            if len(subdomains) == 0:
                err_console.print(f"No subdomains found for {args.domain}")
                return
            console.print(
                f"Found {len(subdomains)} unique subdomain(s) of {args.domain}"
            )
        print_subdomains(subdomains, args.json)
    except CensysRateLimitExceededException:
        err_console.print("Censys API rate limit exceeded")
        print_subdomains(subdomains, args.json)
    except CensysUnauthorizedException:
        err_console.print("Invalid Censys API ID or secret")
        sys.exit(1)
    except CensysException as e:
        err_console.print(str(e))
        sys.exit(1)


def include(parent_parser: argparse._SubParsersAction, parents: dict):
    """Include this subcommand into the parent parser.

    Args:
        parent_parser (argparse._SubParsersAction): Parent parser.
        parents (dict): Parent arg parsers.
    """
    subdomains_parser = parent_parser.add_parser(
        "subdomains",
        description="enumerate subdomains",
        help="enumerate subdomains",
        parents=[parents["auth"]],
    )
    subdomains_parser.add_argument("domain", help="The domain to scan")
    subdomains_parser.add_argument(
        "--max-records", type=int, default=100, help="Max records to query"
    )
    subdomains_parser.add_argument(
        "-j", "--json", action="store_true", help="Output in JSON format"
    )
    subdomains_parser.set_defaults(func=cli_subdomains)



"""An easy-to-use and lightweight API wrapper for Censys Search API (search.censys.io)."""
from .client import SearchClient
from .v1 import CensysCertificates, CensysData
from .v2 import CensysCerts, CensysHosts

__copyright__ = "Copyright 2022 Censys, Inc."
__all__ = [
    "SearchClient",
    "CensysCertificates",
    "CensysData",
    "CensysCerts",
    "CensysHosts",
]



"""Censys search CLI."""
import argparse
import sys
import webbrowser
from typing import Any, Dict, List
from urllib.parse import urlencode

from censys.cli.utils import INDEXES, V1_INDEXES, V2_INDEXES, err_console, write_file
from censys.common.exceptions import CensysCLIException, CensysException
from censys.search import SearchClient
from censys.search.v1.api import CensysSearchAPIv1
from censys.search.v2.api import CensysSearchAPIv2

Results = List[dict]

DEFAULT_FIELDS = {
    "certs": [
        "metadata.updated_at",
        "parsed.issuer.common_name",
        "parsed.names",
        "parsed.serial_number",
        "parsed.self_signed",
        "parsed.subject.common_name",
        "parsed.validity.start",
        "parsed.validity.end",
        "parsed.validity.length",
        "metadata.source",
        "metadata.seen_in_scan",
        "tags",
    ],
}


def cli_search(args: argparse.Namespace):
    """Search subcommand.

    Args:
        args (Namespace): Argparse Namespace.

    Raises:
        CensysCLIException: If invalid options are provided.
    """
    index_type = args.index_type or args.query_type

    if args.open:
        url_query = {"q": args.query}
        if index_type in {"certs", "certificates"}:
            webbrowser.open(
                f"https://search.censys.io/certificates?{urlencode(url_query)}"
            )
            sys.exit(0)
        if index_type in V2_INDEXES:
            url_query.update({"resource": index_type})
            webbrowser.open(f"https://search.censys.io/search?{urlencode(url_query)}")
            sys.exit(0)

    censys_args = {}

    if args.api_id:
        censys_args["api_id"] = args.api_id

    if args.api_secret:
        censys_args["api_secret"] = args.api_secret

    c = SearchClient(**censys_args)

    search_args = {}
    write_args = {"file_format": args.format, "file_path": args.output}
    results: List[Dict[str, Any]] = []

    if index_type in V1_INDEXES:
        v1_index: CensysSearchAPIv1 = getattr(c.v1, index_type)

        if args.max_records:
            search_args["max_records"] = args.max_records

        if args.output:
            if args.output.endswith(".csv"):
                write_args["file_format"] = "csv"
            else:
                write_args["file_format"] = "json"
        else:
            write_args["file_format"] = "screen"

        fields: List[str] = []
        if args.fields:
            if args.overwrite:
                fields = args.fields
            else:
                fields = args.fields + DEFAULT_FIELDS[index_type]
                # Remove duplicates
                fields = list(set(fields))
            write_args["csv_fields"] = fields

        if len(fields) > 20:
            raise CensysCLIException(
                "Too many fields specified. The maximum number of fields is 20."
            )

        search_args["fields"] = fields

        with err_console.status("Searching"):
            try:
                for v1_page in v1_index.search(args.query, **search_args):
                    results.append(v1_page)
            except CensysException:
                err_console.print_exception()
    elif index_type in V2_INDEXES:
        if args.format == "csv" or (args.output and not args.output.endswith(".json")):
            raise CensysCLIException(
                "JSON is the only valid file format for Search 2.0 responses."
            )
        v2_index: CensysSearchAPIv2 = getattr(c.v2, index_type)

        search_args.update(
            {
                "pages": args.pages,
                "per_page": args.per_page,
                "sort": args.sort,
                "virtual_hosts": args.virtual_hosts,
            }
        )
        write_args.update(
            {
                "file_format": "json" if args.output else "screen",
            }
        )

        with err_console.status("Searching"):
            try:
                for v2_page in v2_index.search(args.query, **search_args):
                    results.extend(v2_page)
            except CensysException:
                err_console.print_exception()

    try:
        write_file(results, **write_args)
    except ValueError as error:  # pragma: no cover
        err_console.print(f"Error writing log file. Error: {error}")


def include(parent_parser: argparse._SubParsersAction, parents: dict):
    """Include this subcommand into the parent parser.

    Args:
        parent_parser (argparse._SubParsersAction): Parent parser.
        parents (dict): Parent arg parsers.
    """
    search_parser = parent_parser.add_parser(
        "search",
        description="Query Censys Search for resource data by providing a query \
            string, the resource index, and the fields to be returned",
        help="query Censys search",
        parents=[parents["auth"]],
    )
    search_parser.add_argument(
        "query",
        type=str,
        help="a string written in Censys Search syntax",
    )

    index_metavar = "|".join(INDEXES)
    index_default = "hosts"
    search_parser.add_argument(
        "--index-type",
        type=str,
        default=index_default,
        choices=INDEXES,
        metavar=index_metavar,
        help="which resource index to query",
    )
    # Backwards compatibility
    search_parser.add_argument(
        "--query_type",
        type=str,
        default=index_default,
        choices=INDEXES,
        metavar=index_metavar,
        help=argparse.SUPPRESS,
    )
    search_parser.add_argument(
        "-f",
        "--format",
        type=str,
        default="screen",
        choices=["screen", "json", "csv"],
        metavar="screen|json|csv",
        help=argparse.SUPPRESS,
    )
    search_parser.add_argument(
        "-o",
        "--output",
        type=str,
        help="output file path",
    )
    search_parser.add_argument(
        "-O",
        "--open",
        action="store_true",
        help="open query in browser",
    )

    v2_group = search_parser.add_argument_group(
        f"v2 specific arguments ({', '.join(V2_INDEXES)})"
    )
    v2_group.add_argument(
        "--pages",
        default=1,
        type=int,
        help="number of pages of results to return (when set to -1 returns all pages available)",
    )
    v2_group.add_argument(
        "--per-page",
        default=100,
        type=int,
        help="number of results to return per page",
    )
    v2_group.add_argument(
        "--sort-order",
        dest="sort",
        type=str,
        default="RELEVANCE",
        choices=["RELEVANCE", "ASCENDING", "DESCENDING", "RANDOM"],
        help="sort order of results",
    )
    v2_group.add_argument(
        "--virtual-hosts",
        type=str,
        default="EXCLUDE",
        choices=["INCLUDE", "EXCLUDE", "ONLY"],
        metavar="INCLUDE|EXCLUDE|ONLY",
        help="whether to include virtual hosts in the results",
    )

    v1_group = search_parser.add_argument_group(
        f"v1 specific arguments ({', '.join(V1_INDEXES)})"
    )
    v1_group.add_argument("--fields", nargs="+", help="list of index-specific fields")
    v1_group.add_argument(
        "--overwrite",
        action="store_true",
        default=False,
        help="overwrite instead of append fields returned by default \
            with fields provided in the fields argument",
    )
    v1_group.add_argument(
        "--max-records",
        type=int,
        help="maximum number of results to return",
    )

    search_parser.set_defaults(func=cli_search)



"""Censys HNRI CLI."""
import argparse
import sys
import webbrowser
from typing import Any, List, Optional, Tuple

import requests
from rich import box
from rich.table import Table

from censys.cli.utils import console
from censys.common.exceptions import CensysCLIException, CensysNotFoundException
from censys.search import CensysHosts


class CensysHNRI:
    """Searches the Censys API for the user's current IP to scan for risks."""

    HIGH_RISK_DEFINITION: List[str] = ["TELNET", "REDIS", "POSTGRES", "VNC"]
    MEDIUM_RISK_DEFINITION: List[str] = ["SSH", "HTTP", "HTTPS"]

    def __init__(self, api_id: Optional[str] = None, api_secret: Optional[str] = None):
        """Inits CensysHNRI.

        Args:
            api_id (str): Optional; The API ID provided by Censys.
            api_secret (str): Optional; The API secret provided by Censys.
        """
        self.index = CensysHosts(api_id, api_secret)

    @staticmethod
    def get_current_ip() -> str:
        """Uses ipify.org to get the current IP address.

        Returns:
            str: IP address.
        """
        response = requests.get("https://api.ipify.org?format=json")
        current_ip = str(response.json().get("ip"))
        return current_ip

    def translate_risk(self, services: List[dict]) -> Tuple[List[dict], List[dict]]:
        """Interpret protocols to risks.

        Args:
            services (list): List of services.

        Returns:
            Tuple[list, list]: Lists of high and medium risks.
        """
        high_risk = []
        medium_risk = []

        for service in services:
            service_name = service.get("service_name")
            if service_name in self.HIGH_RISK_DEFINITION:
                high_risk.append(service)
            elif service_name in self.MEDIUM_RISK_DEFINITION:
                medium_risk.append(service)
            else:
                medium_risk.append(service)

        return high_risk, medium_risk

    def make_risks_into_table(self, title: str, risks: List[dict]) -> Table:
        """Creates a table of risks.

        Args:
            title (str): Title of the table.
            risks (list): List of risks.

        Returns:
            Table: Table of risks.
        """
        table = Table("Port", "Service Name", title=title, box=box.SQUARE)
        for risk in risks:
            table.add_row(str(risk.get("port")), risk.get("service_name"))
        return table

    def risks_to_string(self, high_risks: list, medium_risks: list) -> List[Any]:
        """Risks to printable string.

        Args:
            high_risks (list): Lists of high risks.
            medium_risks (list): Lists of medium risks.

        Raises:
            CensysCLIException: No information/risks found.

        Returns:
            list: Printable objects for CLI.
        """
        len_high_risk = len(high_risks)
        len_medium_risk = len(medium_risks)

        if len_high_risk + len_medium_risk == 0:
            raise CensysCLIException

        response: List[Any] = []
        if len_high_risk > 0:
            response.append(
                self.make_risks_into_table(
                    ":exclamation: High Risks Found",
                    high_risks,
                )
            )
        else:
            response.append("You don't have any High Risks in your network\n")
        if len_medium_risk > 0:
            response.append(
                self.make_risks_into_table(
                    ":grey_exclamation: Medium Risks Found",
                    medium_risks,
                )
            )
        else:
            response.append("You don't have any Medium Risks in your network\n")
        return response

    def view_current_ip_risks(self):
        """Gets protocol information for the current IP and returns any risks."""
        current_ip = self.get_current_ip()

        try:
            console.print(f"Searching for information on {current_ip}...")
            results = self.index.view(current_ip)
            services = results.get("services", [])
            high_risk, medium_risk = self.translate_risk(services)
            for res in self.risks_to_string(high_risk, medium_risk):
                console.print(res)
            console.print(
                f"\nFor more information, please visit: https://search.censys.io/hosts/{current_ip}"
            )
        except (CensysNotFoundException, CensysCLIException):
            console.print(
                "[green]:white_check_mark: No Risks were found on your network[/green]"
            )


def cli_hnri(args: argparse.Namespace):
    """HNRI subcommand.

    Args:
        args (Namespace): Argparse Namespace.
    """
    if args.open:
        webbrowser.open("https://search.censys.io/me")
        sys.exit(0)

    client = CensysHNRI(args.api_id, args.api_secret)

    client.view_current_ip_risks()


def include(parent_parser: argparse._SubParsersAction, parents: dict):
    """Include this subcommand into the parent parser.

    Args:
        parent_parser (argparse._SubParsersAction): Parent parser.
        parents (dict): Parent arg parsers.
    """
    hnri_parser = parent_parser.add_parser(
        "hnri",
        description="Home Network Risk Identifier (H.N.R.I.)",
        help="home network risk identifier",
        parents=[parents["auth"]],
    )
    hnri_parser.add_argument(
        "-O",
        "--open",
        action="store_true",
        help="open your IP in browser",
    )
    hnri_parser.set_defaults(func=cli_hnri)



"""Interact with the Censys Seeds, Assets, and Logbook APIs."""
from typing import Optional

from .assets import CertificatesAssets, DomainsAssets, HostsAssets, SubdomainsAssets
from .clouds import Clouds
from .events import Events
from .risks import Risksv1, Risksv2
from .seeds import Seeds


class AsmClient:
    """Client ASM API class."""

    def __init__(self, api_key: Optional[str] = None, **kwargs):
        """Inits AsmClient.

        Args:
            api_key (str): Optional; The API Key provided by Censys.
            **kwargs: Arbitrary keyword arguments.
        """
        self.seeds = Seeds(api_key, **kwargs)
        self.hosts = HostsAssets(api_key, **kwargs)
        self.certificates = CertificatesAssets(api_key, **kwargs)
        self.domains = DomainsAssets(api_key, **kwargs)
        self.events = Events(api_key, **kwargs)
        self.clouds = Clouds(api_key, **kwargs)
        # Risks v1 is approaching deprecation.
        self.risks_v1 = Risksv1(api_key, **kwargs)
        self.risks = self.risks_v1
        self.risks_v2 = Risksv2(api_key, **kwargs)

        # Save the arguments for parameterized client usage
        self.__api_kwargs = kwargs

    def get_subdomains(self, domain: str):
        """Get an API instance for subdomains of the parent domain.

        Args:
            domain: (str): Parent domain to access.

        Returns:
            SubdomainsAssets: A Subdomains Assets API instance .
        """
        return SubdomainsAssets(domain, self.domains._api_key, **self.__api_kwargs)



"""Interact with the Censys Search Certificate API."""
from typing import List

from .api import CensysSearchAPIv1


class CensysCertificates(CensysSearchAPIv1):
    """Interacts with the Certificates index.

    See CensysSearchAPIv1 for additional arguments.

    Args:
        *args: Variable length argument list.
        **kwargs: Arbitrary keyword arguments.
    """

    DEFAULT_URL: str = "https://search.censys.io/api/v1"
    """Default Search API base URL."""
    INDEX_NAME = "certificates"
    """Name of Censys Index."""
    MAX_PER_BULK_REQUEST = 50
    """Max number of bulk requests."""
    bulk_path = f"/bulk/{INDEX_NAME}"

    def bulk(self, fingerprints: List[str]) -> dict:
        """Requests bulk certificates.

        Args:
            fingerprints (List[str]): List of certificate SHA256 fingerprints.

        Returns:
            dict: Search results from an API query.
        """
        result = {}
        start = 0
        end = self.MAX_PER_BULK_REQUEST
        while start < len(fingerprints):
            data = {"fingerprints": fingerprints[start:end]}
            result.update(self._post(self.bulk_path, data=data))
            start = end
            end += self.MAX_PER_BULK_REQUEST

        return result



"""Base for interacting with the Censys Search API."""
import os
from typing import Iterator, List, Optional, Type

from requests.models import Response

from censys.common.base import CensysAPIBase
from censys.common.config import DEFAULT, get_config
from censys.common.exceptions import (
    CensysException,
    CensysExceptionMapper,
    CensysSearchException,
)

Fields = Optional[List[str]]


class CensysSearchAPIv1(CensysAPIBase):
    """This class is the base class for all v1 API indexes."""

    DEFAULT_URL: str = "https://search.censys.io/api/v1"
    """Default Search API base URL."""
    INDEX_NAME: Optional[str] = None
    """Name of Censys Index."""

    def __init__(
        self, api_id: Optional[str] = None, api_secret: Optional[str] = None, **kwargs
    ):
        """Inits CensysSearchAPIv1.

        See CensysAPIBase for additional arguments.

        Args:
            api_id (str): Optional; The API ID provided by Censys.
            api_secret (str): Optional; The API secret provided by Censys.
            **kwargs: Arbitrary keyword arguments.

        Raises:
            CensysException: Base Exception Class for the Censys API.
        """
        CensysAPIBase.__init__(self, kwargs.pop("url", self.DEFAULT_URL), **kwargs)

        # Gets config file
        config = get_config()

        # Try to get credentials
        self._api_id = (
            api_id or os.getenv("CENSYS_API_ID") or config.get(DEFAULT, "api_id")
        )
        self._api_secret = (
            api_secret
            or os.getenv("CENSYS_API_SECRET")
            or config.get(DEFAULT, "api_secret")
        )
        if not self._api_id or not self._api_secret:
            raise CensysException("No API ID or API secret configured.")

        self._session.auth = (self._api_id, self._api_secret)

        # Generate concrete paths to be called
        self.search_path = f"/search/{self.INDEX_NAME}"
        self.view_path = f"/view/{self.INDEX_NAME}/"
        self.report_path = f"/report/{self.INDEX_NAME}"

        # Confirm setup
        # self.account()

    def _get_exception_class(  # type: ignore
        self, res: Response
    ) -> Type[CensysSearchException]:
        return CensysExceptionMapper.SEARCH_EXCEPTIONS.get(
            res.status_code, CensysSearchException
        )

    def account(self) -> dict:
        """Gets the current account information.

        This includes email and quota.

        Returns:
            dict: Account response.
        """
        return self._get("account")

    def quota(self) -> dict:
        """Gets the current account's query quota.

        Returns:
            dict: Quota response.
        """
        return self.account()["quota"]

    def metadata(self, query: str) -> dict:
        """Returns metadata of a given search query.

        Args:
            query (str): The query to be executed.

        Returns:
            dict: The metadata of the result set returned.
        """
        data = {"query": query, "page": 1, "fields": []}
        return self._post(self.search_path, data=data).get("metadata", {})

    def search(
        self,
        query: str,
        fields: Fields = None,
        page: int = 1,
        max_records: Optional[int] = None,
        flatten: bool = True,
    ) -> Iterator[dict]:
        """Searches the given index for all records that match the given query.

        For more details, see our documentation: https://search.censys.io/api

        Args:
            query (str): The query to be executed.
            fields (Fields): Optional; Fields to be returned in the result set.
            page (int): Optional; The page of the result set. Defaults to 1.
            max_records (int): Optional; The maximum number of records.
            flatten (bool): Optional; Flattens fields to dot notation. Defaults to True.

        Raises:
            CensysException: Base Exception Class for the Censys API.

        Yields:
            dict: The result set returned.
        """
        if fields is None:
            fields = []
        try:
            page = int(page)
        except ValueError as error:
            raise CensysException(f"Invalid page value: {page}") from error
        pages = float("inf")
        data = {"query": query, "page": page, "fields": fields, "flatten": flatten}

        count = 0
        while page <= pages:
            payload = self._post(self.search_path, data=data)
            pages = payload["metadata"]["pages"]
            page += 1
            data["page"] = page

            for result in payload["results"]:
                yield result
                count += 1
                if max_records and count >= max_records:
                    return

    def view(self, document_id: str) -> dict:
        """View the current structured data we have on a specific document.

        For more details, see our documentation: https://search.censys.io/api

        Args:
            document_id (str): The ID of the document you are requesting.

        Returns:
            dict: The result set returned.
        """
        return self._get(self.view_path + document_id)

    def report(self, query: str, field: str, buckets: int = 50) -> dict:
        """Creates a report on the breakdown of the values of a field in a result set.

        For more details, see our documentation: https://search.censys.io/api

        Args:
            query (str): The query to be executed.
            field (str): The field you are running a breakdown on.
            buckets (int): Optional; The maximum number of values. Defaults to 50.

        Returns:
            dict: The result set returned.
        """
        data = {"query": query, "field": field, "buckets": int(buckets)}
        return self._post(self.report_path, data=data)



"""Interact with the Censys Search Host API."""
from typing import Any, Dict, List, Optional

from .api import CensysSearchAPIv2
from censys.common.types import Datetime
from censys.common.utils import format_rfc3339


class CensysHosts(CensysSearchAPIv2):
    """Interacts with the Hosts index.

    Examples:
        Inits Censys Hosts.

        >>> from censys.search import CensysHosts
        >>> h = CensysHosts()

        Simple host search.

        >>> for page in h.search("service.service_name: HTTP"):
        >>>     print(page)
        [
            {
            'services':
                [
                    {'service_name': 'HTTP', 'port': 80},
                    {'service_name': 'HTTP', 'port': 443}
                ],
            'ip': '1.0.0.0'
            },
            ...
        ]

        Fetch a specific host and its services

        >>> h.view("1.0.0.0")
        {
            'ip': '8.8.8.8',
            'services': [{}],
            ...
        }

        Simple host aggregate.

        >>> h.aggregate("service.service_name: HTTP", "services.port", num_buckets=5)
        {
            'total_omitted': 591527370,
            'buckets': [
                {'count': 56104072, 'key': '80'},
                {'count': 43527894, 'key': '443'},
                {'count': 23070429, 'key': '7547'},
                {'count': 12970769, 'key': '30005'},
                {'count': 12825150, 'key': '22'}
            ],
            'potential_deviation': 3985101,
            'field': 'services.port',
            'query': 'service.service_name: HTTP',
            'total': 172588754
        }

        Fetch a list of host names for the specified IP address.

        >>> h.view_host_names("1.1.1.1")
        ['one.one.one.one']

        Fetch a list of events for the specified IP address.

        >>> h.view_host_events("1.1.1.1")
        [{'timestamp': '2019-01-01T00:00:00.000Z'}]
    """

    INDEX_NAME = "hosts"
    """Name of Censys Index."""

    def search(
        self,
        query: str,
        per_page: Optional[int] = None,
        cursor: Optional[str] = None,
        pages: int = 1,
        virtual_hosts: Optional[str] = None,
        **kwargs: Any,
    ) -> CensysSearchAPIv2.Query:
        """Search host index.

        Searches the given index for all records that match the given query.
        For more details, see our documentation: https://search.censys.io/api

        Args:
            query (str): The query to be executed.
            per_page (int): Optional; The number of results to be returned for each page. Defaults to 100.
            cursor (int): Optional; The cursor of the desired result set.
            virtual_hosts (str): Optional; Whether to include virtual hosts in the results. Valid values are "EXCLUDE", "INCLUDE", and "ONLY".
            pages (int): Optional; The number of pages returned. Defaults to 1.
            **kwargs (Any): Optional; Additional arguments to be passed to the query.

        Returns:
            Query: Query object that can be a callable or an iterable.
        """
        if virtual_hosts:
            kwargs["virtual_hosts"] = virtual_hosts
        return super().search(query, per_page, cursor, pages, **kwargs)

    def aggregate(
        self,
        query: str,
        field: str,
        num_buckets: Optional[int] = None,
        virtual_hosts: Optional[str] = None,
        **kwargs: Any,
    ) -> dict:
        """Aggregate host index.

        Creates a report on the breakdown of the values of a field in a result set.
        For more details, see our documentation: https://search.censys.io/api

        Args:
            query (str): The query to be executed.
            field (str): The field you are running a breakdown on.
            num_buckets (int): Optional; The maximum number of values. Defaults to 50.
            virtual_hosts (str): Optional; Whether to include virtual hosts in the results. Valid values are "EXCLUDE", "INCLUDE", and "ONLY".
            **kwargs (Any): Optional; Additional arguments to be passed to the query.

        Returns:
            dict: The result set returned.
        """
        if virtual_hosts:
            kwargs["virtual_hosts"] = virtual_hosts
        return super().aggregate(query, field, num_buckets, **kwargs)

    def view_host_names(
        self, ip: str, per_page: Optional[int] = None, cursor: Optional[str] = None
    ) -> List[str]:
        """Fetches a list of host names for the specified IP address.

        Args:
            ip (str): The IP address of the requested host.
            per_page (int): Optional; The number of results to be returned for each page. Defaults to 100.
            cursor (int): Optional; The cursor of the desired result set.

        Returns:
            List[str]: A list of host names.
        """
        args = {"per_page": per_page, "cursor": cursor}
        return self._get(self.view_path + ip + "/names", args)["result"]["names"]

    def view_host_diff(
        self,
        ip: str,
        ip_b: Optional[str] = None,
        at_time: Optional[Datetime] = None,
        at_time_b: Optional[Datetime] = None,
    ):
        """Fetches a diff of the specified IP address.

        Args:
            ip (str): The IP address of the requested host.
            ip_b (str): Optional; The IP address of the second host.
            at_time (Datetime): Optional; An RFC3339 timestamp which represents
                the point-in-time used as the basis for Host A.
            at_time_b (Datetime): Optional; An RFC3339 timestamp which represents
                the point-in-time used as the basis for Host B.

        Returns:
            dict: A diff of the hosts.
        """
        args: Dict[str, Any] = {}
        if ip_b:
            args["ip_b"] = ip_b
        if at_time:
            args["at_time"] = format_rfc3339(at_time)
        if at_time_b:
            args["at_time_b"] = format_rfc3339(at_time_b)
        return self._get(self.view_path + ip + "/diff", args)["result"]

    def view_host_events(
        self,
        ip: str,
        start_time: Optional[Datetime] = None,
        end_time: Optional[Datetime] = None,
        per_page: Optional[int] = None,
        cursor: Optional[str] = None,
        reversed: Optional[bool] = None,
    ) -> List[dict]:
        """Fetches a list of events for the specified IP address.

        Args:
            ip (str): The IP address of the requested host.
            start_time (Datetime): Optional; An RFC3339 timestamp which represents
                the beginning chronological point-in-time (inclusive) from which events are returned.
            end_time (Datetime): Optional; An RFC3339 timestamp which represents
                the ending chronological point-in-time (exclusive) from which events are returned.
            per_page (int): Optional; The maximum number of hits to return in each response
                (minimum of 1, maximum of 50).
            cursor (str): Optional; Cursor token from the API response.
            reversed (bool): Optional; Reverse the order of the return events,
                that is, return events in reversed chronological order.

        Returns:
            List[dict]: A list of events.
        """
        args = {"per_page": per_page, "cursor": cursor, "reversed": reversed}
        if start_time:
            args["start_time"] = format_rfc3339(start_time)
        if end_time:
            args["end_time"] = format_rfc3339(end_time)

        return self._get(f"/v2/experimental/{self.INDEX_NAME}/{ip}/events", args)[
            "result"
        ]["events"]

    def list_hosts_with_tag(self, tag_id: str) -> List[str]:
        """Returns a list of hosts which are tagged with the specified tag.

        Args:
            tag_id (str): The ID of the tag.

        Returns:
            List[str]: A list of host IP addresses.
        """
        hosts = self._list_documents_with_tag(tag_id, "hosts", "hosts")
        return [host["ip"] for host in hosts]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['censys',
 'censys.asm',
 'censys.asm.assets',
 'censys.asm.risks',
 'censys.cli',
 'censys.cli.commands',
 'censys.common',
 'censys.search',
 'censys.search.v1',
 'censys.search.v2']

package_data = \
{'': ['*']}

install_requires = \
['backoff>=2.0.1,<3.0.0', 'requests>=2.26.0', 'rich>=10.16.2']

extras_require = \
{':python_version < "3.8"': ['importlib-metadata']}

entry_points = \
{'console_scripts': ['censys = censys.cli:main']}

setup_kwargs = {
    'name': 'censys',
    'version': '2.1.7',
    'description': 'An easy-to-use and lightweight API wrapper for Censys APIs (censys.io).',
    'long_description': '# Censys Python Library\n\n[![PyPI](https://img.shields.io/pypi/v/censys?color=orange&logo=pypi&logoColor=orange)](https://pypi.org/project/censys/)\n[![Python Version](https://img.shields.io/badge/python-3.7%2B-blue?logo=python)](https://www.python.org/downloads/)\n[![Read the Docs (version)](https://img.shields.io/readthedocs/censys-python/latest?logo=read%20the%20docs)](https://censys-python.readthedocs.io/en/stable/?badge=stable)\n[![GitHub Discussions](https://img.shields.io/badge/GitHub-Discussions-brightgreen?logo=github)](https://github.com/censys/censys-python/discussions)\n[![PRs Welcome](https://img.shields.io/badge/PRs-welcome-organge.svg?logo=git&logoColor=organge)](http://makeapullrequest.com)\n[![License](https://img.shields.io/github/license/censys/censys-python?logo=apache)](https://github.com/censys/censys-python/blob/main/LICENSE)\n\nAn easy-to-use and lightweight API wrapper for Censys APIs ([censys.io](https://censys.io/)). Python 3.7+ is currently supported.\n\n> **Notice:** The Censys Search v1 endpoints are deprecated as of Nov. 30, 2021. Please begin using v2 endpoints to query hosts and certificates and check out our [support center](https://support.censys.io/hc/en-us/sections/360013076551-Censys-Search-2-0) for resources.\n\n## Features\n\n- [Search Censys data](https://censys-python.readthedocs.io/en/stable/usage-v2.html)\n- [Bulk Certificate lookups](https://censys-python.readthedocs.io/en/stable/usage-v1.html#bulk)\n- [Download Bulk Data](https://censys-python.readthedocs.io/en/stable/usage-v1.html#data)\n- [Manage assets, events, and seeds in Censys ASM](https://censys-python.readthedocs.io/en/stable/usage-asm.html)\n- [Command-line interface](https://censys-python.readthedocs.io/en/stable/cli.html)\n\n<a href="https://asciinema.org/a/500416" target="_blank"><img src="https://asciinema.org/a/500416.svg" width="600"/></a>\n\n## Getting Started\n\nThe library can be installed using `pip`.\n\n```sh\npip install censys\n```\n\nTo upgraded using `pip`.\n\n```sh\npip install --upgrade censys\n```\n\nTo configure your search credentials run `censys config` or set both `CENSYS_API_ID` and `CENSYS_API_SECRET` environment variables.\n\n```sh\n$ censys config\n\nCensys API ID: XXX\nCensys API Secret: XXX\nDo you want color output? [y/n]: y\n\nSuccessfully authenticated for your@email.com\n```\n\nTo configure your ASM credentials run `censys asm config` or set the `CENSYS_ASM_API_KEY` environment variables.\n\n```sh\n$ censys asm config\n\nCensys ASM API Key: XXX\nDo you want color output? [y/n]: y\n\nSuccessfully authenticated\n```\n\n## API Reference and User Guide available on [Read the Docs](https://censys-python.readthedocs.io/)\n\n[![Read the Docs](https://raw.githubusercontent.com/censys/censys-python/main/docs/_static/readthedocs.png)](https://censys-python.readthedocs.io/)\n\n## Resources\n\n- [Source](https://github.com/censys/censys-python)\n- [Issue Tracker](https://github.com/censys/censys-python/issues)\n- [Changelog](https://github.com/censys/censys-python/releases)\n- [Documentation](https://censys-python.rtfd.io)\n- [Discussions](https://github.com/censys/censys-python/discussions)\n- [Censys Homepage](https://censys.io/)\n- [Censys Search](https://search.censys.io/)\n\n## Contributing\n\nAll contributions (no matter how small) are always welcome. See [Contributing to Censys Python](.github/CONTRIBUTING.md)\n\n## Development\n\nThis project uses [poetry](https://python-poetry.org/) for dependency management. Please ensure you have [installed the latest version](https://python-poetry.org/docs/#installation).\n\n```sh\ngit clone git@github.com:censys/censys-python.git\ncd censys-python/\npoetry install\n```\n\n## Testing\n\n```sh\n# Run tests\npoetry run pytest\n# With coverage report\npoetry run pytest --cov-report html\n```\n\n## License\n\nThis software is licensed under [Apache License, Version 2.0](http://www.apache.org/licenses/LICENSE-2.0)\n\n- Copyright (C) 2022 Censys, Inc.\n',
    'author': 'Censys, Inc.',
    'author_email': 'support@censys.io',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



"""Censys config CLI."""
import argparse
import os
import sys

from rich.prompt import Confirm, Prompt

from censys.cli.utils import console
from censys.common.config import DEFAULT, get_config, write_config
from censys.common.exceptions import CensysUnauthorizedException
from censys.search.v2.api import CensysSearchAPIv2


def cli_config(_: argparse.Namespace):  # pragma: no cover
    """Config subcommand.

    Args:
        _: Argparse Namespace.
    """
    api_id_prompt = "Censys API ID"
    api_secret_prompt = "Censys API Secret"

    config = get_config()
    api_id = config.get(DEFAULT, "api_id")
    api_secret = config.get(DEFAULT, "api_secret")

    api_id_env = os.getenv("CENSYS_API_ID")
    api_secret_env = os.getenv("CENSYS_API_SECRET")

    if api_id_env is not None or api_secret_env is not None:
        console.print(
            "Please note environment variables (CENSYS_API_ID & CENSYS_API_SECRET) "
            "will take priority over configured credentials."
        )
        api_id = api_id_env or api_id
        api_secret = api_secret_env or api_secret

    if api_id and api_secret:
        redacted_id = api_id.replace(api_id[:32], 32 * "*")
        redacted_secret = api_secret.replace(api_secret[:28], 28 * "*")
        api_id_prompt = f"{api_id_prompt} [cyan]({redacted_id})[/cyan]"
        api_secret_prompt = f"{api_secret_prompt} [cyan]({redacted_secret})[/cyan]"

    api_id = Prompt.ask(api_id_prompt, console=console) or api_id
    api_secret = Prompt.ask(api_secret_prompt, console=console) or api_secret

    if not (api_id and api_secret):
        console.print("Please enter valid credentials")
        sys.exit(1)

    api_id = api_id.strip()
    api_secret = api_secret.strip()

    color = Confirm.ask(
        "Do you want color output?", default=True, show_default=False, console=console
    )
    config.set(DEFAULT, "color", "auto" if color else "")

    try:
        client = CensysSearchAPIv2(api_id, api_secret)
        account = client.account()
        email = account.get("email")
        console.print(f"\nSuccessfully authenticated for {email}")

        # Assumes that login was successfully
        config.set(DEFAULT, "api_id", api_id)
        config.set(DEFAULT, "api_secret", api_secret)

        write_config(config)
        sys.exit(0)
    except CensysUnauthorizedException:
        console.print("Failed to authenticate")
        sys.exit(1)
    except PermissionError as e:
        console.print(e)
        console.print(
            "Cannot write config file to directory. "
            + "Please set the `CENSYS_CONFIG_PATH` environmental variable to a writeable location."
        )
        sys.exit(1)


def include(parent_parser: argparse._SubParsersAction, parents: dict):
    """Include this subcommand into the parent parser.

    Args:
        parent_parser (argparse._SubParsersAction): Parent parser.
        parents (dict): Parent arg parsers.
    """
    config_parser = parent_parser.add_parser(
        "config",
        description="Configure Censys Search API Settings",
        help="configure Censys search API settings",
    )
    config_parser.set_defaults(func=cli_config)



"""Interact with the config file."""
import configparser
import os
from pathlib import Path

DEFAULT = "DEFAULT"
HOME_PATH = str(Path.home())
CENSYS_PATH = os.path.join(HOME_PATH, ".config", "censys")
CONFIG_PATH = os.path.join(CENSYS_PATH, "censys.cfg")

default_config = {
    "api_id": "",
    "api_secret": "",
    "asm_api_key": "",
    "color": "auto",
}


def get_config_path() -> str:
    """Returns the path to the config file.

    Returns:
        str: Path to config file.
    """
    alt_path = os.getenv("CENSYS_CONFIG_PATH")
    if alt_path:
        return alt_path
    return CONFIG_PATH


def write_config(config: configparser.ConfigParser) -> None:
    """Writes config to file.

    Args:
        config (configparser.ConfigParser): Configuration to write.

    Raises:
        PermissionError: If the config file is not writable.
    """
    config_path = get_config_path()
    if config_path == CONFIG_PATH:
        if not os.access(HOME_PATH, os.W_OK):
            raise PermissionError(
                "Cannot write to home directory. Please set the `CENSYS_CONFIG_PATH` environmental variable to a writeable location."
            )
        elif not os.path.isdir(CENSYS_PATH):
            os.makedirs(CENSYS_PATH)
    with open(config_path, "w") as configfile:
        config.write(configfile)


def get_config() -> configparser.ConfigParser:
    """Reads and returns config.

    Returns:
        configparser.ConfigParser: Config for Censys.
    """
    config = configparser.ConfigParser(defaults=default_config, default_section=DEFAULT)
    config_path = get_config_path()
    if os.path.isfile(config_path):
        config.read(config_path)
    return config



"""Interact with the Censys Assets API."""
from .assets import Assets
from .certificates import CertificatesAssets
from .domains import DomainsAssets
from .hosts import HostsAssets
from .subdomains import SubdomainsAssets

__all__ = [
    "Assets",
    "CertificatesAssets",
    "DomainsAssets",
    "HostsAssets",
    "SubdomainsAssets",
]



"""Interact with the Censys Risks v1 API."""
from typing import Iterator, Optional

from ..api import CensysAsmAPI


class Risksv1(CensysAsmAPI):
    """Risks v1 API class."""

    base_path = "/v1/risks"

    def get_risks(
        self,
        cloud: Optional[str] = None,
        environment: Optional[str] = None,
        include_accepted_risks: Optional[bool] = None,
        page_number: int = 1,
        page_size: int = 100,
    ) -> Iterator[dict]:
        """Retrieve risks.

        Returns a full list of all risks that affect any assets in the system,
        along with a count of assets affected by each risk.

        Args:
            cloud (str): The cloud to filter by.
            environment (str): The environment to filter by.
            include_accepted_risks (bool): Whether to include accepted risks.
            page_number (int): The page number to return.
            page_size (int): The number of items to return per page.

        Yields:
            dict: Host count result.
        """
        yield from self._get_page(
            self.base_path,
            page_number=page_number,
            page_size=page_size,
            args={
                "cloud": cloud,
                "environment": environment,
                "includeAcceptedRisks": include_accepted_risks,
            },
            keyword="data",
        )



#!/usr/bin/env python3
"""Interact with the Censys Search API through the command line."""
import sys

from .args import get_parser
from censys.common.version import __version__


def main():
    """Main cli function."""
    parser = get_parser()

    # Executes by subcommand
    args = parser.parse_args()

    if args.version:
        print(f"Censys Python Version: {__version__}")
        sys.exit(0)

    try:
        args.func(args)
    except KeyboardInterrupt:  # pragma: no cover
        sys.exit(1)


if __name__ == "__main__":  # pragma: no cover
    main()



"""Interact with argparser."""
import argparse
import os

from . import commands
from censys.common.config import DEFAULT, get_config


def get_parser() -> argparse.ArgumentParser:
    """Gets ArgumentParser for CLI.

    Returns:
        argparse.ArgumentParser
    """
    config = get_config()

    auth = argparse.ArgumentParser(add_help=False)
    auth.add_argument(
        "--api-id",
        default=os.getenv("CENSYS_API_ID") or config.get(DEFAULT, "api_id"),
        required=False,
        help="a Censys API ID \
            (alternatively you can use the env variable CENSYS_API_ID)",
    )
    auth.add_argument(
        "--api-secret",
        default=os.getenv("CENSYS_API_SECRET") or config.get(DEFAULT, "api_secret"),
        required=False,
        help="a Censys API SECRET \
            (alternatively you can use the env variable CENSYS_API_SECRET)",
    )
    asm_auth = argparse.ArgumentParser(add_help=False)
    asm_auth.add_argument(
        "--api-key",
        default=os.getenv("CENSYS_ASM_API_KEY") or config.get(DEFAULT, "asm_api_key"),
        required=False,
        help="a Censys ASM API Key \
            (alternatively you can use the env variable CENSYS_ASM_API_KEY)",
    )

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-v",
        "--version",
        action="store_true",
        default=False,
        help="display version",
    )

    def print_help(_: argparse.Namespace):
        """Prints help."""
        parser.print_help()
        parser.exit()

    parser.set_defaults(func=print_help)

    subparsers = parser.add_subparsers()

    parents = {
        "auth": auth,
        "asm_auth": asm_auth,
    }
    for command in commands.__dict__.values():
        try:
            include_func = command.include
        except AttributeError:
            continue

        include_func(subparsers, parents)

    return parser
